
from ._dynamicTree import cutreeHybrid
import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
from sklearn import linear_model
from scipy.stats import norm
from scipy import stats
import networkx as nx
import datetime
import seaborn as sns
from scipy.cluster import hierarchy  
from scipy import cluster   
from sklearn import decomposition as skldec 
from matplotlib.colors import LinearSegmentedColormap
import scanpy as sc
from typing import Union,Tuple
import matplotlib


from scipy.spatial.distance import pdist
from scipy.cluster.hierarchy import linkage,dendrogram
from ..utils import pyomic_palette,plot_network
import os


class pyWGCNA(object):
    r"""
        pyWGCNA: Weighted correlation network analysis in Python
    """
    def __init__(self,data:pd.DataFrame,save_path:str=''):
        r"""Initialize the pyWGCNA module
        
        Arguments:
            data: The dataframe of gene expression data
            save_path: The path to save the results
        """
        self.data=data.fillna(0)
        self.data_len=len(data)
        self.data_index=data.index
        self.save_path=save_path

    def calculate_correlation_direct(self,method:str='pearson',save:bool=False):
        """calculate the correlation coefficient matrix

        Arguments:
            method: The method to calculate the correlation coefficient matrix
            save: Whether to save the result
        
        """
        print('...correlation coefficient matrix is being calculated')
        self.result=self.data.T.corr(method)
        if save==True:
            self.result.to_csv(self.save_path+'/'+'direction correlation matrix.csv')
            print("...direction correlation have been saved")


    def calculate_correlation_indirect(self,save:bool=False):
        """calculate the indirect correlation coefficient matrix
        
        Arguments:
            save: Whether to save the result
        """
        print('...indirect correlation matrix is being calculated')
        np.fill_diagonal(self.result.values, 0)
        arr=abs(self.result)
        arr_len=len(arr)
        self.temp=np.zeros(arr_len)
        for i in range(1,3):    
            self.temp=self.temp+(arr**i)/i

        from sklearn import preprocessing
        min_max_scaler = preprocessing.MinMaxScaler()
        X_minMax = min_max_scaler.fit_transform(self.temp)
        self.normal_corelation=pd.DataFrame(X_minMax,index=self.temp.index,columns=self.temp.columns)

        if save==True:
            self.temp.to_csv(self.save_path+'/'+'indirection correlation matrix.csv')
            print("...indirection correlation have been saved")


    def calculate_soft_threshold(self,threshold_range:int=12,
                                 plot:bool=True,save:bool=False,
                                 figsize:tuple=(6,3))->pd.DataFrame:
        """calculate the soft threshold
        
        Arguments:
            threshold_range: The range of threshold
            plot: Whether to plot the result
            save: Whether to save the result

        Returns:
            The dataframe of soft threshold
        """

        print('...soft_threshold is being calculated')
        soft=6
        re1=pd.DataFrame(columns=['beta','r2','meank'])
        for j in range(1,threshold_range):
            #print('Now is:',j)
            result_i=np.float_power(self.temp,j)
            tt_0=np.sum(abs(result_i),axis=0)-1
            n=np.histogram(tt_0), #
            x=n[0][0]
            y=[]
            for i in range(len(n[0][1])-1):
                y.append((n[0][1][i]+n[0][1][i+1])/2)
            x=np.log10(x)
            y=np.log10(y)
            res=stats.linregress(x, y)
            r2=np.float_power(res.rvalue,2)
            k=tt_0.mean()
            re1.loc[j]={'beta':j,'r2':r2,'meank':k}
        for i in re1['r2']:
            if i>0.85:
                soft=re1[re1['r2']==i]['beta'].iloc[0]
                break
        self.re1=re1
        self.soft=soft
        print('...appropriate soft_thresholds:',soft)
        if plot==True:
            fig, ax = plt.subplots(1,2,figsize=figsize)
            ax[0].scatter(re1['beta'],re1['r2'],c='b')
            ax[0].plot([0,threshold_range],[0.95,0.95],c='r')
            ax[0].set_xlim(0,threshold_range)
            ax[0].set_ylabel('r2',fontsize=14)
            ax[0].set_xlabel('beta',fontsize=14)
            ax[0].set_title('Best Soft threshold')

            ax[1].scatter(re1['beta'],re1['meank'],c='r')
            ax[1].set_ylabel('meank',fontsize=14)
            ax[1].set_xlabel('beta',fontsize=14)
            ax[1].set_title('Best Soft threshold')
            fig.tight_layout()
        if save==True:
            fig.savefig(self.save_path+'/'+'soft_threshold_hist.png',dpi=300)
        return re1



    def calculate_corr_matrix(self):
        """calculate the correlation matrix
        
        """
        self.cor=np.float_power(self.temp,self.soft)
        np.fill_diagonal(self.cor.values, 1.0)

    def calculate_distance(self,trans:bool=True):
        """calculate the distance matrix
        
        Arguments:
            trans: Whether to transpose the correlation matrix
        """
        #distance
        print("...distance have being calculated")
        if trans==True:
            self.distances = pdist(1-self.cor, "euclidean")
        else:
            self.distances = pdist(self.cor, "euclidean")

    def calculate_geneTree(self,linkage_method:str='ward'):
        """
        calculate the geneTree
        
        Arguments:
            linkage_method: The method to calculate the geneTree, it can be found in `scipy.cluster.hierarchy.linkage`
        """

        #geneTree
        print("...geneTree have being calculated")
        self.geneTree=linkage(self.distances, linkage_method)

    def calculate_dynamicMods(self,minClusterSize:int=30,
                  deepSplit:int=2,):
        """calculate the dynamicMods

        Arguments:
            minClusterSize: The minimum size of cluster
            deepSplit: The deep of split
        """

        #dynamicMods
        print("...dynamicMods have being calculated")
        self.dynamicMods=cutreeHybrid(self.geneTree,distM=self.distances,minClusterSize = minClusterSize,
                                            deepSplit = deepSplit, pamRespectsDendro = False)
        print("...total:",len(set(self.dynamicMods['labels'])))

    def calculate_gene_module(self,figsize:tuple=(25,10),save:bool=True,
                              colorlist:list=None)->pd.DataFrame:
        """calculate the gene module
        
        Arguments:
            figsize: The size of figure
            save: Whether to save the figure
            colorlist: The color list of module

        Returns:
            The dataframe of gene module
        """
        plt.figure(figsize=figsize)
        grid = plt.GridSpec(3, 1, wspace=0.5, hspace=0.1)

        plt.subplot(grid[0:2,0])
        hierarchy.set_link_color_palette(['#000000'])
        dn=hierarchy.dendrogram(self.geneTree,color_threshold=0, above_threshold_color='black')
        plt.tick_params( \
            axis='x',
            which='both',
            bottom='off',
            top='off',
            labelbottom='off')
        
        #mol
        #ivl=dn['ivl']
        mod=[self.dynamicMods['labels'][int(x)] for x in dn['ivl']]
        plot_mod=np.array([mod])
        gene_name=[self.cor.index[int(x)] for x in dn['ivl']]
        mol=pd.DataFrame(columns=['ivl','module','name'])
        mol['ivl']=dn['ivl']
        mol['module']=mod
        mol['name']=gene_name

        res=[]
        for i in plot_mod[0]:
            if i in res:
                continue
            else:
                res.append(i)
        res_len=len(res)
        if colorlist!=None:
            colorlist_cmap=LinearSegmentedColormap.from_list('Custom', colorlist[:res_len], len(colorlist[:res_len]))
            color_dict=dict(zip(range(1,res_len+1),colorlist[:res_len]))
            mol['color']=mol['module'].map(color_dict)
            plt.subplot(grid[2,0])
            ax1=plt.pcolor(plot_mod,cmap=colorlist_cmap)
        else:
            if res_len>28:
                colorlist=sc.pl.palettes.default_102
            else:
                colorlist=pyomic_palette()
            colorlist_cmap=LinearSegmentedColormap.from_list('Custom', colorlist[:res_len], len(colorlist[:res_len]))
            color_dict=dict(zip(range(1,res_len+1),colorlist[:res_len]))
            mol['color']=mol['module'].map(color_dict)
            plt.subplot(grid[2,0])
            ax1=plt.pcolor(plot_mod,cmap=colorlist_cmap)

        self.mol=mol
        if save==True:
            plt.savefig(self.save_path+'/'+'module_tree.png',dpi=300)
        return mol
    
    def get_sub_module(self,mod_list:list)->pd.DataFrame:
        '''
        Get sub-module of a module

        Arguments:
            mod_list: module number

        Returns:
            sub_module:sub-module of a module
        '''
        mol=self.mol
        return mol[mol['module'].isin(mod_list)]
    
    def get_sub_network(self,mod_list:list,correlation_threshold=0.95)->nx.Graph:
        '''
        Get sub-network of a module

        Arguments:
            mod_list: module number

        Returns:
            sub_network:sub-network of a module
        '''
        module1=self.get_sub_module(mod_list)
        gene_net1=abs(self.normal_corelation.loc[module1.name.tolist(),module1.name.tolist()])


        G = nx.Graph()
        for i in gene_net1.index:
            for j in gene_net1.columns:
                if gene_net1.loc[i,j]>correlation_threshold:
                    G.add_weighted_edges_from([(i, j, gene_net1.loc[i,j])])
        return G
    
    def plot_matrix(self,cmap='RdBu_r',save:bool=True,figsize:tuple=(8,9),
                    legene_ncol:int=2,legene_bbox_to_anchor:tuple=(5, 2.95),legene_fontsize:int=12,):
        """plot the matrix of correlation

        Arguments:
            cmap: The color of matrix
            save: Whether to save the figure
            figsize: The size of figure
            legene_ncol: The number of column of legene
            legene_bbox_to_anchor: The position of legene
            legene_fontsize: The size of legene

        Returns:
            ax: The axis of figure
        """

        module_color=self.mol.copy()
        module_color.index=module_color['name']
        a=sns.clustermap(self.normal_corelation,#standard_scale=1,
                    cmap=cmap,yticklabels=False,xticklabels=False,#method='ward',metric='euclidean',
                    #row_cluster=False,col_cluster=False,
                    col_linkage=self.geneTree,row_linkage=self.geneTree,
                    col_colors=module_color.loc[self.normal_corelation.index,'color'].values,
                    row_colors=module_color.loc[self.normal_corelation.index,'color'].values,
                    cbar_kws={"shrink": .5},square=True,
                    dendrogram_ratio=(.1, .2),
                    cbar_pos=(-.1, .32, .03, .2),figsize=figsize
        )
        #add the color of the module
        for i in range(len(set(self.mol['module']))):
            t1=self.mol.loc[self.mol['module']==i+1]
            a.ax_heatmap.add_patch(plt.Rectangle((t1.index[0],t1.index[0]),len(t1),len(t1),fill=False,color='black',lw=1))

        color=[]
        labels=[]
        for i in module_color['module'].unique():
            t1=module_color.loc[module_color['module']==i].iloc[0]
            color.append(t1['color'])
            labels.append(i)
        from matplotlib import patches as mpatches
        #用label和color列表生成mpatches.Patch对象，它将作为句柄来生成legend
        patches = [ mpatches.Patch(color=color[i], label="ME-{}".format(labels[i]) ) for i in range(len(labels)) ] 
        plt.legend(handles=patches,bbox_to_anchor=legene_bbox_to_anchor, ncol=legene_ncol,fontsize=legene_fontsize)
        if save==True:
            plt.savefig(self.save_path+'/'+'module_matrix.png',dpi=300,bbox_inches = 'tight')
        return a

    
    def plot_sub_network(self,mod_list:list,
                         correlation_threshold:float=0.95,
                         plot_genes=None,
                         plot_gene_num:int=5,**kwargs)->Tuple[matplotlib.figure.Figure,matplotlib.axes._axes.Axes]:
        '''
        plot sub-network of a module

        Arguments:
            mod_list: module number
            correlation_threshold: correlation threshold
            plot_genes: genes to plot in the sub-network. If None, the hub genes will be ploted
            plot_gene_num: number of genes to plot

        Returns:
            fig: figure
            ax: axis
        '''

        module1=self.get_sub_module(mod_list)
        G_type_dict=dict(zip(module1['name'],[str(i) for i in module1['module']]))
        G_color_dict=dict(zip(module1['name'],module1['color']))
        G=self.get_sub_network(mod_list,correlation_threshold)
        G_color_dict=dict(zip(G.nodes,[G_color_dict[i] for i in G.nodes]))
        G_type_dict=dict(zip(G.nodes,[G_type_dict[i] for i in G.nodes]))

        degree_dict = dict(G.degree(G.nodes()))
        de_pd=pd.DataFrame(degree_dict.values(),index=degree_dict.keys(),columns=['Degree'])
        hub_gene=[]
        if plot_genes!=None:
            hub_gene=plot_genes
        else:
            for i in mod_list:
                ret_gene=list(set(de_pd.index) & set(module1.loc[module1['module']==i]['name'].tolist()))
                hub_gene1=de_pd.loc[ret_gene,'Degree'].sort_values(ascending=False)[:plot_gene_num].index.tolist()
                hub_gene+=hub_gene1

        fig,ax=plot_network(G,G_type_dict,G_color_dict,plot_node=hub_gene,**kwargs)
        return fig,ax
        
    def analysis_meta_correlation(self,meta_data:pd.DataFrame)->Tuple[pd.DataFrame,pd.DataFrame]:
        """Analysis meta correlation
        
        Arguments:
            meta_data: meta data of samples

        Returns:
            meta_cor: meta correlation
            meta_p: meta p-value
        """

        print("...PCA analysis have being done")
        data=self.data
        module=self.mol
        pcamol=pd.DataFrame(columns=data.columns)
        set_index=set(module['module'])
        for j in set_index:
            newdata=pd.DataFrame(columns=data.columns)
            for i in list(module[module['module']==j].dropna()['name']):
                newdata.loc[i]=data[data.index==i].values[0]
            from sklearn.decomposition import PCA
            pca = PCA(n_components=1) 
            reduced_X = pca.fit_transform(newdata.T)
            tepcamol=pd.DataFrame(reduced_X.T,columns=data.columns)
            pcamol.loc[j]=tepcamol.values[0]
        pcamol.index=set_index
        
        print("...co-analysis have being done")
        from scipy.stats import spearmanr,pearsonr,kendalltau
        # seed random number generator
        # calculate spearman's correlation
        new_meta=pd.DataFrame()
        new_meta.index=meta_data.index
        for j in meta_data.columns:
            if meta_data[j].dtype=='int64':
                new_meta=pd.concat([new_meta,meta_data[j]],axis=1)
            elif meta_data[j].dtype!='float32': 
                new_meta=pd.concat([new_meta,meta_data[j]],axis=1)
            elif meta_data[j].dtype!='float64':
                new_meta=pd.concat([new_meta,meta_data[j]],axis=1)
            else:
                one_hot = pd.get_dummies(meta_data[j], prefix=j)
                new_meta=pd.concat([new_meta,one_hot],axis=1)
                

        result_1=pd.DataFrame(columns=new_meta.columns)
        result_p=pd.DataFrame(columns=new_meta.columns)
        for j in new_meta.columns:
            co=[]
            pvv=[]
            for i in range(len(pcamol)):   
                tempcor=pd.DataFrame(columns=['x','y'])
                tempcor['x']=list(new_meta[j])
                tempcor['y']=list(pcamol.iloc[i])
                tempcor=tempcor.dropna()
                coef,pv=pearsonr(tempcor['x'],tempcor['y'])
                co.append(coef)
                pvv.append(pv)
            result_1[j]=co
            result_p[j]=pvv
                #print(coef)
        result_1=abs(result_1)
        result_1.index=set_index

        return result_1,result_p
    
    def plot_meta_correlation(self,cor_matrix:tuple,
                              label_fontsize:int=10,label_colors:str='red')->matplotlib.axes._axes.Axes:
        """Plot meta correlation
        
        Arguments:
            cor_matrix: meta correlation and meta p-value
            label_fontsize: label fontsize
            label_colors: label colors

        Returns:
            ax: axis
        """

        cor=cor_matrix[0].copy()
        cor_p=cor_matrix[1].copy()
        cor['module']=cor.index
        cor_p['module']=cor_p.index

        df_melt = pd.melt(cor, id_vars='module', var_name='meta', value_name='correlation')
        df_melt2 = pd.melt(cor_p, id_vars='module', var_name='meta', value_name='pvalue')
        df_melt['pvalue']=df_melt2['pvalue']
        df_melt['logp']=-np.log10(df_melt2['pvalue'])
        df_melt.index=[str(i) for i in df_melt.index]
        df_melt['module']=[str(i) for i in df_melt['module']]

        #new_keys = {'item_key': 'module','group_key': 'meta','sizes_key': 'logp','color_key': 'correlation'}
        
        import dotplot
        new_keys = {'item_key': 'meta','group_key': 'module','sizes_key': 'logp','color_key': 'correlation'}
        dp = dotplot.DotPlot.parse_from_tidy_data(df_melt, **new_keys)
        #fig, ax = plt.subplots(figsize=(10,2))
        ax = dp.plot(size_factor=10, cmap='Spectral_r',vmax=1,vmin=0,
                    dot_title = '–log10(Pvalue)', colorbar_title = 'Correlation',)
        xlabs=ax.get_axes()[0].get_xticklabels()
        ax.get_axes()[0].set_xticklabels(xlabs,fontsize=label_fontsize)
        ylabs=ax.get_axes()[0].get_yticklabels()
        ax.get_axes()[0].set_yticklabels(ylabs,fontsize=label_fontsize)
        #去除配体
        [i.set_text(i.get_text().split('_')[1]) for i in xlabs]
        ax.get_axes()[0].set_xticklabels(xlabs,fontsize=label_fontsize)

        #上色        
        ax.get_axes()[0].tick_params(axis='both',colors=label_colors, which='both')
        return ax






    def __Analysis_cocharacter(self,
                        character,save=True):

        '''
        Calculate gene and trait correlation matrix

        Parameters
        ----------
        - character: `pandas.DataFrame`
            DataFrame of sample's character, columns=character, index=['sample1','sample2',...]
        - save: `bool`
            save the result

        Returns
        -------
        - result: `pandas.DataFrame`
            Character and module correlation matrix
        '''
        print("...PCA analysis have being done")
        data=self.data
        module=self.mol
        pcamol=pd.DataFrame(columns=data.columns)
        set_index=set(module['module'])
        for j in set_index:
            newdata=pd.DataFrame(columns=data.columns)
            for i in list(module[module['module']==j].dropna()['name']):
                newdata.loc[i]=data[data.index==i].values[0]
            from sklearn.decomposition import PCA
            pca = PCA(n_components=1) 
            reduced_X = pca.fit_transform(newdata.T)
            tepcamol=pd.DataFrame(reduced_X.T,columns=data.columns)
            pcamol.loc[j]=tepcamol.values[0]
        pcamol.index=set_index
        
        print("...co-analysis have being done")
        from scipy.stats import spearmanr,pearsonr,kendalltau
        # seed random number generator
        # calculate spearman's correlation
        result_1=pd.DataFrame(columns=character.columns)
        result_p=pd.DataFrame(columns=character.columns)
        for j in character.columns:
            co=[]
            pvv=[]
            for i in range(len(pcamol)):   
                tempcor=pd.DataFrame(columns=['x','y'])
                tempcor['x']=list(character[j])
                tempcor['y']=list(pcamol.iloc[i])
                tempcor=tempcor.dropna()
                coef,pv=pearsonr(tempcor['x'],tempcor['y'])
                co.append(coef)
                pvv.append(pv)
            result_1[j]=co
            result_p[j]=pvv
                #print(coef)
        result_1=abs(result_1)
        result_1.index=set_index
        
        plt.figure(figsize=(10,10))
        sns.heatmap(result_1,vmin=0, vmax=1,center=1,annot=True,square=True)
        if save==True:
            plt.savefig(self.save_path+'/'+'co_character.png',dpi=300)
        return result_1



def Trans_corr_matrix(data,
                    method='pearson',
                    cmap='seismic'):
    '''
    Calculate the correlation adjacent matrix (direct and indirect) (scale-free network)

    Parameters
    ----------
    - data: `pandas.DataFrame`
        DataFrame of data points with each entry in the form:[sample1','sample2'...],index=gene_name
    - method: `str`
        The method of calculating correlation coefficient
        method='pearson'/'kendall'/'spearman'
    - cmap: `str`
        color Style of drawing

    Returns
    ----------
    - result: `pandas.DataFrame`
        the gene's correlation adjacent matrix
    '''

    data_len=len(data)
    data_index=data.index
    
    #correlation coefficient
    start = datetime.datetime.now()
    print('...correlation coefficient matrix is being calculated')
    result=data.T.corr(method)
    end = datetime.datetime.now()
    result.to_csv('direction correlation matrix.csv')
    print("...direction correlation have been saved")
    print("...calculate time",end-start)
    
    #indirect correlation add
    start = datetime.datetime.now()
    print('...indirect correlation matrix is being calculated')
    np.fill_diagonal(result.values, 0)
    arr=abs(result)
    arr_len=len(arr)
    temp=np.zeros(arr_len)
    for i in range(1,3):    
        temp=temp+(arr**i)/i
    end = datetime.datetime.now()
    temp.to_csv('indirection correlation matrix.csv')
    print("...indirection correlation have been saved")
    print("...calculate time",end-start)
    
        
    #cal soft_threshold
    plt.rc('font', family='Arial')
    plt.rcParams["font.weight"] = "bold"
    plt.rcParams["font.weight"] = "12"
    plt.rcParams["axes.labelweight"] = "bold"
    my_dpi=300
    fig=plt.figure(figsize=(2000/my_dpi, 1000/my_dpi), dpi=my_dpi)
    start = datetime.datetime.now()
    print('...soft_threshold is being calculated')
    soft=6
    re1=pd.DataFrame(columns=['beta','r2','meank'])
    for j in range(1,12):
        result_i=np.float_power(temp,j)
        tt_0=np.sum(abs(result_i),axis=0)-1
        n=plt.hist(x = tt_0), #
        x=n[0][0]
        y=[]
        for i in range(len(n[0][1])-1):
            y.append((n[0][1][i]+n[0][1][i+1])/2)
        x=np.log10(x)
        y=np.log10(y)
        res=stats.linregress(x, y)
        r2=np.float_power(res.rvalue,2)
        k=tt_0.mean()
        re1=re1.append({'beta':j,'r2':r2,'meank':k},ignore_index=True)
    for i in re1['r2']:
        if i>0.85:
            soft=re1[re1['r2']==i]['beta'].iloc[0]
            break
    print('...appropriate soft_thresholds:',soft)
    plt.savefig('soft_threshold_hist.png',dpi=300)
    
    #select soft_threhold
    my_dpi=300
    fig=plt.figure(figsize=(2000/my_dpi, 1000/my_dpi), dpi=my_dpi)
    grid = plt.GridSpec(1, 4, wspace=1, hspace=0.1)
    #fig, (ax0, ax1) = plt.subplots(2, 1)
    plt.subplot(grid[0,0:2])
    cmap1=sns.color_palette(cmap)
    plt.subplot(1,2, 1)
    p1=sns.regplot(x=re1["beta"], y=re1['r2'], fit_reg=False, marker="o", color=cmap1[0])
    p1.axhline(y=0.9,ls=":",c=cmap1[5])

    plt.subplot(grid[0,2:4])
    p1=sns.regplot(x=re1["beta"], y=re1['meank'], fit_reg=False, marker="o", color=cmap1[4])
    plt.savefig('soft_threshold_select.png',dpi=300)
    
    test=np.float_power(temp,soft)
    np.fill_diagonal(test.values, 1.0)
    return test
  
def Select_module(data,
                  linkage_method='ward',
                  minClusterSize=30,
                  deepSplit=2,
                  cmap='seismic',
                  trans=True,
                 ):

    '''
    Select gene module from correlation adjacent matrix

    Parameters
    ----------
    - data: `pandas.DataFrame`
        DataFrame of data points with each entry in the form:[sample1','sample2'...],index=gene_name
    - linkage_method: `str`
        The method of clustering
        method='ward'/'average'...
    - minClusterSize: `int`
        The least contained gene in each module
    - deepSplit: `int`
        Degree of split
    - cmap: `str`
        color style of figure

    Returns
    ----------
    - mol: `pandas.DataFrame`
        the gene's module
        columns=['ivl','module','name']
    '''

    #distance
    print("...distance have being calculated")
    if trans==True:
       distances = pdist(1-data, "euclidean")
    else:
        distances = pdist(data, "euclidean")
    
    #geneTree
    print("...geneTree have being calculated")
    geneTree=linkage(distances, linkage_method)
      
    #dynamicMods
    print("...dynamicMods have being calculated")
    dynamicMods=dynamicTree.cutreeHybrid(geneTree,distM=distances,minClusterSize = minClusterSize,deepSplit = deepSplit, pamRespectsDendro = False)
    #return dynamicMods


    print("...total:",len(set(dynamicMods['labels'])))
    
    plt.rc('font', family='Arial')
    plt.rcParams["font.weight"] = "bold"
    plt.rcParams["font.weight"] = "12"
    plt.rcParams["axes.labelweight"] = "bold"
    plt.figure(figsize=(25, 10))
    grid = plt.GridSpec(3, 1, wspace=0.5, hspace=0.1)
    #fig, (ax0, ax1) = plt.subplots(2, 1)
    plt.subplot(grid[0:2,0])
    hierarchy.set_link_color_palette(['#000000'])
    dn=hierarchy.dendrogram(geneTree,color_threshold=0, above_threshold_color='black')
    plt.tick_params( \
        axis='x',
        which='both',
        bottom='off',
        top='off',
        labelbottom='off')
    
    #mol
    x=dn['ivl']
    y=[dynamicMods['labels'][int(x)] for x in dn['ivl']]
    yy=np.array([y])
    z=[data.index[int(x)] for x in dn['ivl']]
    mol=pd.DataFrame(columns=['ivl','module','name'])
    mol['ivl']=x
    mol['module']=y
    mol['name']=z
    
    plt.subplot(grid[2,0])
    ax1=plt.pcolor(yy,cmap=cmap)

    plt.savefig('module_tree.png',dpi=300)
    return mol
     

    

def Analysis_cocharacter(data,
                        character,
                        module):

    '''
    Calculate gene and trait correlation matrix

    Parameters
    ----------
    - data: `pandas.DataFrame`
        DataFrame of data points with each entry in the form:[sample1','sample2'...],index=gene_name
    - character: `pandas.DataFrame`
        DataFrame of sample's character, columns=character, index=['sample1','sample2',...]
    - module: `pandas.DataFrame`
        The result of the function Select_module

    Returns
    ----------
    - result: `pandas.DataFrame`
        Character and module correlation matrix
    '''
    print("...PCA analysis have being done")
    pcamol=pd.DataFrame(columns=data.columns)
    set_index=set(module['module'])
    for j in set_index:
        newdata=pd.DataFrame(columns=data.columns)
        for i in list(module[module['module']==j].dropna()['name']):
            newdata=newdata.append(data[data.index==i])
        from sklearn.decomposition import PCA
        pca = PCA(n_components=1) 
        reduced_X = pca.fit_transform(newdata.T)
        tepcamol=pd.DataFrame(reduced_X.T,columns=data.columns)
        pcamol=pcamol.append(tepcamol,ignore_index=True)
    pcamol.index=set_index
    
    print("...co-analysis have being done")
    from scipy.stats import spearmanr,pearsonr,kendalltau
    # seed random number generator
    # calculate spearman's correlation
    result_1=pd.DataFrame(columns=character.columns)
    result_p=pd.DataFrame(columns=character.columns)
    for j in character.columns:
        co=[]
        pvv=[]
        for i in range(len(pcamol)):   
            tempcor=pd.DataFrame(columns=['x','y'])
            tempcor['x']=list(character[j])
            tempcor['y']=list(pcamol.iloc[i])
            tempcor=tempcor.dropna()
            coef,pv=pearsonr(tempcor['x'],tempcor['y'])
            co.append(coef)
            pvv.append(pv)
        result_1[j]=co
        result_p[j]=pvv
            #print(coef)
    result_1=abs(result_1)
    result_1.index=set_index
    
    plt.rc('font', family='Arial')
    plt.rcParams["font.weight"] = "bold"
    plt.rcParams["font.weight"] = "12"
    plt.rcParams["axes.labelweight"] = "bold"
    plt.figure(figsize=(10,10))
    sns.heatmap(result_1,vmin=0, vmax=1,center=1,annot=True,square=True)
    plt.savefig('co_character.png',dpi=300)
    return result_1


from __future__ import division
import numpy.ma as ma
import numpy as np
import scipy as s
import math

from ..utils import *
from .. import gpu_utils

from time import time

# Import manually defined functions
from .variational_nodes import BernoulliGaussian_Unobserved_Variational_Node
from .variational_nodes import UnivariateGaussian_Unobserved_Variational_Node
from .variational_nodes import UnivariateGaussian_Unobserved_Variational_Node_with_MultivariateGaussian_Prior

class W_Node(UnivariateGaussian_Unobserved_Variational_Node):
    def __init__(self, dim, pmean, pvar, qmean, qvar, qE=None, qE2=None):
        super().__init__(dim=dim, pmean=pmean, pvar=pvar, qmean=qmean, qvar=qvar, qE=qE, qE2=qE2)

    def precompute(self, options):
        # Precompute terms to speed up computation
        gpu_utils.gpu_mode = options['gpu_mode']

    def removeFactors(self, idx):
        super().removeFactors(idx, axis=1)

    def updateParameters(self, ix=None, ro=1.):
        """
        Public function to update the nodes parameters
        Optional arguments for stochastic updates are:
            - ix: list of indices of the minibatch
            - ro: step size of the natural gradient ascent
        """
        #-----------------------------------------------------------------------
        # get Expectations which are necessarry for the update
        #-----------------------------------------------------------------------
        Y = self.markov_blanket["Y"].get_mini_batch()
        mask = self.markov_blanket["Y"].getMask()
        Z = self.markov_blanket["Z"].get_mini_batch()
        tau = self.markov_blanket["Tau"].get_mini_batch()
        tau[mask] = 0.

        # Collect parameters from the prior or expectations from the markov blanket
        Mu = self.P.getParameters()["mean"]
        if "AlphaW" in self.markov_blanket:
            Alpha = self.markov_blanket['AlphaW'].getExpectation(expand=True)
        else:
            Alpha = 1./self.P.params['var']

        # Collect parameters from the Q distributions of this node
        Q = self.Q.getParameters()
        Qmean, Qvar = Q['mean'], Q['var']

        # compute stochastic "anti-bias" coefficient
        N = self.markov_blanket["Y"].dim[0]
        coeff = float(N) / float(Y.shape[0])

        # make sure ro is not None
        if ro is None:
            ro = 1.

        # compute the update
        self._updateParameters(Y, Z, tau, Mu, Alpha, Qmean, Qvar, coeff, ro)

    def _updateParameters(self, Y, Z, tau, Mu, Alpha, Qmean, Qvar, coeff, ro):

        for k in range(self.dim[1]):
            foo = coeff * np.dot(Z["E2"][:,k],tau)

            bar_tmp1 = gpu_utils.array(Z["E"][:,k])

            bar_tmp2 = - gpu_utils.dot(gpu_utils.array(Z["E"][:,np.arange(self.dim[1])!=k]),
                               gpu_utils.array(Qmean[:,np.arange(self.dim[1])!=k].T))
            bar_tmp2 += gpu_utils.array(Y)
            bar_tmp2 *= gpu_utils.array(tau)

            bar = coeff * gpu_utils.asnumpy(gpu_utils.dot(bar_tmp1, bar_tmp2))

            # stochastic update of W
            Qvar[:,k] *= (1 - ro)
            Qvar[:,k] += ro/(Alpha[:,k]+foo)

            # NOTE Do not use "Qvar" in the update like we used to because this
            # does not hold for stochastic because of the ro weighting
            Qmean[:,k] *= (1 - ro)
            Qmean[:,k] += ro * (1/(Alpha[:,k]+foo)) * (bar + Alpha[:,k]*Mu[:,k])

    def calculateELBO(self):

        # Collect parameters and expectations of current node
        Qpar,Qexp = self.Q.getParameters(), self.Q.getExpectations()
        Qmean, Qvar = Qpar['mean'], Qpar['var']
        QE, QE2 = Qexp['E'],Qexp['E2']

        if "MuW" in self.markov_blanket:
            PE, PE2 = self.markov_blanket['MuW'].getExpectations()['E'], self.markov_blanket['MuW'].getExpectations()['E2']
        else:
            PE, PE2 = self.P.getParameters()["mean"], np.zeros((self.dim[0],self.dim[1]))

        if 'AlphaW' in self.markov_blanket:
            Alpha = self.markov_blanket["AlphaW"].getExpectations(expand=True)
        else:
            Alpha = dict()
            Alpha['E'] = 1./self.P.params['var']
            Alpha['lnE'] = np.log(1./self.P.params['var'])

        # compute term from the exponential in the Gaussian
        tmp1 = 0.5 * QE2 - PE * QE + 0.5 * PE2
        tmp1 = -(tmp1 * Alpha['E']).sum()

        # compute term from the precision factor in front of the Gaussian
        tmp2 = 0.5 * Alpha["lnE"].sum()

        lb_p = tmp1 + tmp2
        lb_q = -(np.log(Qvar).sum() + self.dim[0]*self.dim[1])/2.

        return lb_p-lb_q


class SW_Node(BernoulliGaussian_Unobserved_Variational_Node):
    def __init__(self, dim, pmean_S0, pmean_S1, pvar_S0, pvar_S1, ptheta, qmean_S0, qmean_S1, qvar_S0, qvar_S1, qtheta, qEW_S0=None, qEW_S1=None, qES=None):
        super().__init__(dim, pmean_S0, pmean_S1, pvar_S0, pvar_S1, ptheta, qmean_S0, qmean_S1, qvar_S0, qvar_S1, qtheta, qEW_S0, qEW_S1, qES)

    def precompute(self, options):
        self.factors_axis = 1
        gpu_utils.gpu_mode = options['gpu_mode']

    def removeFactors(self, idx):
        super().removeFactors(idx, axis=1)

    def updateParameters(self, ix=None, ro=1.):

        # Collect expectations from other nodes
        Y = self.markov_blanket["Y"].get_mini_batch()
        Z = self.markov_blanket["Z"].get_mini_batch()
        tau = self.markov_blanket["Tau"].get_mini_batch()
        mask = self.markov_blanket["Y"].getMask()

        if "AlphaW" in self.markov_blanket:
            Alpha = self.markov_blanket["AlphaW"].getExpectation(expand=True)
        else:
            Alpha = 1./self.P.params['var_B1']

        thetatmp = self.markov_blanket["ThetaW"].getExpectations(expand=True)
        theta_lnE, theta_lnEInv  = thetatmp['lnE'], thetatmp['lnEInv']

        # This node
        SW = self.Q.getExpectations()["E"]
        Q = self.Q.getParameters()
        Qmean_S1, Qvar_S1, Qvar_S0 = Q['mean_B1'], Q['var_B1'],  Q['var_B0']
        Qtheta = Q['theta']

        # Compute stochastic "anti-bias" coefficient
        N = self.markov_blanket["Y"].dim[0]
        coeff = float(N) / float(Y.shape[0])

        # Compute parameter updates
        self._updateParameters(Y, Z, tau, mask, Alpha, Qmean_S1, Qvar_S1, Qvar_S0, Qtheta, SW, theta_lnE, theta_lnEInv, coeff, ro)
    
    def _updateParameters(self, Y, Z, tau, mask, Alpha, Qmean_S1, Qvar_S1, Qvar_S0, Qtheta, SW, theta_lnE, theta_lnEInv, coeff, ro):

        # Mask matrices
        tau[mask] = 0.

        # Copy matrices to GPU
        # Y_gpu = gpu_utils.array(Y)
        tau_gpu = gpu_utils.array(tau)
        Z_gpu = gpu_utils.array(Z["E"])
        ZZ_gpu = gpu_utils.array(Z["E2"])
        # precompute terms
        # tauY_gpu = gpu_utils.array(tau*Y).T
        tauY_gpu = (tau_gpu*gpu_utils.array(Y)).T
        
        foo = gpu_utils.asnumpy( gpu_utils.dot(ZZ_gpu.T, tau_gpu).T )
        term4_tmp1 = gpu_utils.asnumpy( gpu_utils.dot(tauY_gpu, Z_gpu) )

        del tauY_gpu, ZZ_gpu

        # Update each latent variable in turn
        for k in range(self.dim[1]):

            # Compute terms
            term1 = (theta_lnE-theta_lnEInv)[:,k]

            term2 = 0.5*np.log(Alpha[:,k])
            term3 = 0.5 * coeff * np.log(foo[:,k] + Alpha[:,k])

            # term4_tmp1 = gpu_utils.dot(tauYT, Zk_cp)

            # term4_tmp2_1 = gpu_utils.array(SW[:,np.arange(self.dim[1])!=k].T)
            # term4_tmp2_2 = (Z_gpu[:,k] * gpu_utils.array(Z['E'][:,np.arange(self.dim[1])!=k]).T).T
            # term4_tmp2 = (tau_gpu*gpu_utils.dot(term4_tmp2_2, term4_tmp2_1)).sum(axis=0)
            term4_tmp2 = gpu_utils.asnumpy( (tau_gpu*gpu_utils.dot(
                (Z_gpu[:,k] * gpu_utils.array(Z['E'][:,np.arange(self.dim[1])!=k]).T).T, 
                gpu_utils.array(SW[:,np.arange(self.dim[1])!=k].T))
            ).sum(axis=0) )

            term4_tmp3 = foo[:,k] + Alpha[:,k]

            term4 = coeff * 0.5*np.divide(np.square(term4_tmp1[:,k]-term4_tmp2),term4_tmp3)

            # Update S
            Qtheta[:,k] *= (1 - ro)
            Qtheta[:,k] += ro * (1./(1.+np.exp(-(term1+term2-term3+term4))))

            # Update W
            tmp_var = 1./term4_tmp3
            Qvar_S1[:,k] *= (1 - ro)
            Qvar_S1[:,k] += ro * tmp_var

            Qmean_S1[:,k] *= (1 - ro)
            Qmean_S1[:,k] += ro * tmp_var * (term4_tmp1[:,k]-term4_tmp2)

            # Update Expectations for the next iteration
            SW[:,k] = Qtheta[:,k] * Qmean_S1[:,k]

            del term1, term2, term3, term4_tmp2, term4_tmp3

        # update of Qvar_S0
        Qvar_S0 *= (1 - ro)
        Qvar_S0 += ro/Alpha

        # Save updated parameters of the Q distribution
        self.Q.setParameters(mean_B0=np.zeros((self.dim[0],self.dim[1])), var_B0=Qvar_S0, mean_B1=Qmean_S1, var_B1=Qvar_S1, theta=Qtheta)

    def calculateELBO(self):
        # Collect parameters and expectations
        Qpar,Qexp = self.Q.getParameters(), self.Q.getExpectations()
        S,WW = Qexp["EB"], Qexp["ENN"]
        Qvar = Qpar['var_B1']
        theta = self.markov_blanket["ThetaW"].getExpectations(expand=True)
        if "AlphaW" in self.markov_blanket:
            alpha = self.markov_blanket["AlphaW"].getExpectations(expand=True)
        else:
            alpha = dict()
            alpha['E'] = 1./self.P.params['var_B1']
            alpha['lnE'] = np.log(1./self.P.params['var_B1'])

        # Calculate ELBO term for W
        lb_pw = (alpha["lnE"].sum() - np.sum(alpha["E"]*WW))/2.
        lb_qw = -0.5*self.dim[1]*self.dim[0] - 0.5*(S*np.log(Qvar) + (1.-S)*np.log(1./alpha["E"])).sum()
        lb_w = lb_pw - lb_qw

        # Calculate ELBO term for S
        # S[S<1e-10] = 1e-10
        # S[S>0.9999999] = 0.9999999
        lb_ps = S*theta['lnE'] + (1.-S)*theta['lnEInv']
        lb_qs = S*np.log(S) + (1.-S)*np.log(1.-S)

        # Replace NAs (due to theta=1) with zeros
        lb_ps[np.isnan(lb_ps)] = 0.
        lb_qs[np.isnan(lb_qs)] = 0.

        lb_s = np.sum(lb_ps) - np.sum(lb_qs)

        return lb_w + lb_s



import numpy as np
import scipy as s
from .basic_distributions import Distribution

from ... import config


class Bernoulli(Distribution):
    """
    Class to define Bernoulli distributions

    Equations:

    """
    def __init__(self, dim, theta, E=None):
        Distribution.__init__(self, dim)

        # Initialise parameters
        theta = np.ones(dim) * theta
        self.params = { 'theta':theta }

        # Initialise expectations
        if E is None:
            self.updateExpectations()
        else:
            self.expectations = { 'E':np.ones(dim)*E }

        # float64 -> float32
        if config.use_float32: self.to_float32()

        # Check that dimensionalities match
        self.CheckDimensionalities()

    def updateExpectations(self):
        E = self.params['theta']
        self.expectations = { 'E':E }

    def density(self, x):
        assert x.shape == self.dim, "Problem with the dimensionalities"
        return np.prod( self.params['theta']**x * (1-self.params['theta'])**(1-x) )

    def loglik(self, x):
        assert x.shape == self.dim, "Problem with the dimensionalities"
        return np.sum( x*np.log(self.params['theta']) + (1-x)*np.log(1-self.params['theta']) )

    def sample(self):
        return np.random.binomial(1, self.params['theta'])



import os 
import scanpy as sc
import pandas as pd
import numpy as np
from ._SCSA import Process,Annotator
import sys
import argparse
import gzip
import time
import requests
import anndata
from ..pp._preprocess import scale
from ..utils import gen_mpl_labels

def global_imports(modulename,shortname = None, asfunction = False):
    if shortname is None: 
        shortname = modulename
    if asfunction is False:
        globals()[shortname] = __import__(modulename)
    else:        
        globals()[shortname] = __import__(modulename)

metatime_install=False
def check_metatime():
    """
    
    """
    global metatime_install
    try:
        import metatime
        metatime_install=True
        print('metatime have been install version:',metatime.__version__)
    except ImportError:
        raise ImportError(
            'Please install the metatime: `pip install metatime`.'
        )


def data_downloader(url,path,title):
    r"""datasets downloader
    
    Parameters
    ----------
    - url: `str`
        the download url of datasets
    - path: `str`
        the save path of datasets
    - title: `str`
        the name of datasets
    
    Returns
    -------
    - path: `str`
        the save path of datasets
    """
    if os.path.isfile(path):
        print("......Loading dataset from {}".format(path))
        return path
    else:
        print("......Downloading dataset save to {}".format(path))
        
    dirname, _ = os.path.split(path)
    try:
        if not os.path.isdir(dirname):
            print("......Creating directory {}".format(dirname))
            os.makedirs(dirname, exist_ok=True)
    except OSError as e:
        print("......Unable to create directory {}. Reason {}".format(dirname,e))
    
    
    start = time.time()
    size = 0
    res = requests.get(url, stream=True)

    chunk_size = 102400
    content_size = int(res.headers["content-length"]) 
    if res.status_code == 200:
        print('......[%s Size of file]: %0.2f MB' % (title, content_size/chunk_size/10.24))
        with open(path, 'wb') as f:
            for data in res.iter_content(chunk_size=chunk_size):
                f.write(data)
                size += len(data) 
                print('\r'+ '......[Downloader]: %s%.2f%%' % ('>'*int(size*50/content_size), float(size/content_size*100)), end='')
        end = time.time()
        print('\n' + ".......Finish！%s.2f s" % (end - start))
    
    return path


def data_preprocess(adata,clustertype='leiden',
                    path='temp/rna.csv',layer='scaled',rank_rep=False):
    r"""data preprocess for SCSA
    
    Parameters
    ----------
    - adata: `AnnData`
        AnnData object
    - path: `str`   
        the save path of datasets

    Returns
    -------
    - adata: `AnnData`
        AnnData object
    """
    dirname, _ = os.path.split(path)
    try:
        if not os.path.isdir(dirname):
            print("......Creating directory {}".format(dirname))
            os.makedirs(dirname, exist_ok=True)
    except OSError as e:
        print("......Unable to create directory {}. Reason {}".format(dirname,e))

    sc.settings.verbosity = 2  # reduce the verbosity
    if rank_rep==False and 'rank_genes_groups' not in adata.uns.keys():
        sc.tl.rank_genes_groups(adata, clustertype, method='wilcoxon')
    else:
        sc.tl.rank_genes_groups(adata, clustertype, method='wilcoxon')
    result = adata.uns['rank_genes_groups']
    groups = result['names'].dtype.names
    dat = pd.DataFrame({group + '_' + key[:1]: result[key][group] for group in groups for key in ['names', 'logfoldchanges','scores','pvals']})
    dat.to_csv(path)
    return dat

def __cell_annotate(data,
                foldchange=1.5,pvalue=0.05,
                output='temp/rna_anno.txt',
                outfmt='txt',Gensymbol=True,
                species='Human',weight=100,tissue='All',
                celltype='normal',norefdb=False,noprint=True,list_tissue=False):
    r"""cell annotation by SCSA
    
    Parameters
    ----------
    - data: `AnnData`
        AnnData object
    - foldchange: `float`
        foldchange threshold
    - pvalue: `float`
        pvalue threshold
    - output: `str`
        the save path of annotation result
    - outfmt: `str`
        the format of annotation result
    - Gensymbol: `bool`
        whether to use gene symbol
    - species: `str`
        the species of datasets
    - weight: `int`
        the weight of datasets
    - tissue: `str`
        the tissue of datasets
    - celltype: `str`
        the celltype of datasets
    - norefdb: `bool`
        whether to use reference database
    - noprint: `bool`
        whether to print the result
    - list_tissue: `bool`
        whether to list the tissue of datasets
    
    Returns
    -------
    - result: `pandas.DataFrame`
        the annotation result
    """
    data.to_csv('temp/rna.csv')

    #https://figshare.com/ndownloader/files/37262710
    model_path = data_downloader(url='https://figshare.com/ndownloader/files/37262710',
                                    path='temp/whole.db',title='whole')
    print('......Auto annotate cell')
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('-i', '--input', default = "temp/rna.csv")
    parser.add_argument('-o', '--output',default=output)
    parser.add_argument('-d', '--db', default = model_path,)
    parser.add_argument('-s', '--source', default = "scanpy",)
    parser.add_argument('-c', '--cluster', default = "all",)
    parser.add_argument('-f',"--fc",default = "1.5",)
    parser.add_argument('-fc',"--foldchange",default =foldchange,)
    parser.add_argument('-p',"--pvalue",default = pvalue,)
    parser.add_argument('-w',"--weight",default = weight,)
    parser.add_argument('-g',"--species",default = species,)
    parser.add_argument('-k',"--tissue",default = tissue,)
    parser.add_argument('-m', '--outfmt', default = outfmt, )
    parser.add_argument('-T',"--celltype",default = celltype,)
    parser.add_argument('-t', '--target', default = "cellmarker",)
    parser.add_argument('-E',"--Gensymbol",action = "store_true",default=Gensymbol,)
    parser.add_argument('-N',"--norefdb",action = "store_true",default=norefdb,)
    parser.add_argument('-b',"--noprint",action = "store_true",default=noprint,)
    parser.add_argument('-l',"--list_tissue",action = "store_true",default = False,)
    parser.add_argument('-M', '--MarkerDB',)
    args = parser.parse_args()

    p = Process()
    if args.list_tissue:
        p.list_tissue(args)
    p.run_cmd(args)

    result=pd.read_csv('temp/rna_anno.txt',sep='\t')
    return result


def __cell_anno_print(anno):
    r"""print the annotation result
    
    Parameters
    ----------
    - anno: `pandas.DataFrame`
        the annotation result
    Returns
    -------

    """
    for i in set(anno['Cluster']):
        test=anno.loc[anno['Cluster']==i].iloc[:2]
        if test.iloc[0]['Z-score']>test.iloc[1]['Z-score']*2:
            print('Nice:Cluster:{}\tCell_type:{}\tZ-score:{}'.format(i,test.iloc[0]['Cell Type'],
                                                        np.around(test.iloc[0]['Z-score'],3)))
        else:
            print('Cluster:{}\tCell_type:{}\tZ-score:{}'.format(i,('|').join(test['Cell Type'].values.tolist()),
                                                        ('|').join(np.around(test['Z-score'].values,3).astype(str).tolist())))

def scanpy_lazy(adata:anndata.AnnData,min_genes:int=200,min_cells:int=3,drop_doublet:bool=True,
                n_genes_by_counts:int=4300,pct_counts_mt:int=25,
                target_sum:float=1e4,min_mean:float=0.0125, max_mean:int=3, min_disp:float=0.5,max_value:int=10,
                n_comps:int=100, svd_solver:str="auto",
                n_neighbors:int=15, random_state:int = 112, n_pcs:int=50,
                )->anndata.AnnData:
    r"""scanpy lazy analysis
    
    Arguments:
        adata: AnnData object
        min_genes: the min number of genes
        min_cells: the min number of cells
        drop_doublet: whether to drop doublet
        n_genes_by_counts: the max number of genes
        pct_counts_mt: the max proportion of mito-genes
        target_sum: the max counts of total_counts
        min_mean: the min mean of genes
        max_mean: the max mean of genes
        min_disp: the min dispersion of genes
        max_value: the max value of genes
        n_comps: the number of components
        svd_solver: the solver of svd
        n_neighbors: the number of neighbors
        random_state: the random state
        n_pcs: the number of pcs

    Returns:
        adata: AnnData object
    """
    #filter cells and genes
    sc.pp.filter_cells(adata, min_genes=min_genes)
    sc.pp.filter_genes(adata, min_cells=min_cells)
    #filter the doublets cells
    if drop_doublet:
        sc.external.pp.scrublet(adata) #estimates doublets
        adata = adata[adata.obs['predicted_doublet'] == False] #do the actual filtering
    #calculate the proportion of mito-genes
    adata.var['mt'] = adata.var_names.str.startswith('MT-')  # annotate the group of mitochondrial genes as 'mt'
    sc.pp.calculate_qc_metrics(adata, qc_vars=['mt'], percent_top=None, log1p=False, inplace=True)
    adata = adata[adata.obs.n_genes_by_counts < n_genes_by_counts, :]
    adata = adata[adata.obs.pct_counts_mt < pct_counts_mt, :]
    #save the raw counts
    adata.layers["counts"] = adata.X.copy()
    #normalization, the max counts of total_counts is 20000 means the amount is 10e4
    sc.pp.normalize_total(adata, target_sum=target_sum)
    #log
    sc.pp.log1p(adata)
    #select high-variable genes
    sc.pp.highly_variable_genes(adata, min_mean=min_mean, max_mean=max_mean, min_disp=min_disp)
    #save and filter
    adata.raw = adata
    adata = adata[:, adata.var.highly_variable]
    #scale
    #scale(adata, max_value=max_value)
    sc.pp.scale(adata, max_value=max_value)
    #pca analysis
    sc.tl.pca(adata, n_comps=n_comps, svd_solver=svd_solver)
    #pca(adata,layer='scaled',n_pcs=50)
    #cell neighbors graph construct
    sc.pp.neighbors(adata, n_neighbors=n_neighbors, random_state = random_state, n_pcs=n_pcs)
    #umap
    sc.tl.leiden(adata)
    sc.tl.paga(adata)
    sc.pl.paga(adata, plot=False)  # remove `plot=False` if you want to see the coarse-grained graph
    sc.tl.umap(adata, init_pos='paga')
    return adata

def scanpy_cellanno_from_dict(adata:anndata.AnnData,
                               anno_dict:dict,
                               anno_name:str='major',
                               clustertype:str='leiden',
                               )->None:
    r"""add cell type annotation from dict to anndata object

    Arguments:
        adata: AnnData object of scRNA-seq after preprocessing
        anno_dict: dict of cell type annotation. key is the cluster name, value is the cell type name.like `{'0':'B cell','1':'T cell'}`
        anno_name: the name of annotation
        clustertype: Clustering name used in scanpy. (leiden)

    """

    adata.obs[anno_name+'_celltype'] = adata.obs[clustertype].map(anno_dict).astype('category')
    print('...cell type added to {}_celltype on obs of anndata'.format(anno_name))

def get_celltype_marker(adata:anndata.AnnData,
                            clustertype:str='leiden',
                            log2fc_min:int=2,scores_type='scores',
                            pval_cutoff:float=0.05,rank:bool=False)->dict:
        r"""Get marker genes for each clusters.
        
        Arguments:
            adata: anndata object
            clustertype: Clustering name used in scanpy. (leiden)
            log2fc_min: Minimum log2 fold change of marker genes. (2)
            pval_cutoff: Maximum p value of marker genes. (0.05)
            rank: Whether to rank genes by wilcoxon test. (True)
            scores_type: The type of scores. can be selected from `scores` and `logfoldchanges`

        Returns:
            cellmarker: A dictionary of marker genes for each clusters.
        """
        print('...get cell type marker')
        celltypes = sorted(adata.obs[clustertype].unique())
        cell_marker_dict={}
        if rank==False:
            sc.tl.rank_genes_groups(adata, clustertype, method='wilcoxon')
        for celltype in celltypes:
            degs = sc.get.rank_genes_groups_df(adata, group=celltype, key='rank_genes_groups', log2fc_min=log2fc_min, 
                                            pval_cutoff=pval_cutoff)
            foldp=np.histogram(degs[scores_type])
            foldchange=(foldp[1][np.where(foldp[1]>0)[0][-5]]+foldp[1][np.where(foldp[1]>0)[0][-6]])/2
            
            cellmarker=degs.loc[degs[scores_type]>foldchange]['names'].values
            cell_marker_dict[celltype]=cellmarker

        return cell_marker_dict






class pySCSA(object):

    def __init__(self,adata:anndata.AnnData,
                foldchange:float=1.5,pvalue:float=0.05,
                output:str='temp/rna_anno.txt',
                model_path:str='',
                outfmt:str='txt',Gensymbol:bool=True,
                species:str='Human',weight:int=100,tissue:str='All',target:str='cellmarker',
                celltype:str='normal',norefdb:bool=False,cellrange:str=None,
                noprint:bool=True,list_tissue:bool=False) -> None:

        r"""Initialize the pySCSA class

        Arguments:
            adata: AnnData object of scRNA-seq after preprocessing
            foldchange: Fold change threshold for marker filtering. (2.0)
            pvalue: P-value threshold for marker filtering. (0.05)
            output: Output file for marker annotation.(temp/rna_anno.txt)
            model_path: Path to the Database for annotation. If not provided, the model will be downloaded from the internet.
            outfmt: Output format for marker annotation. (txt)
            Gensymbol: Using gene symbol ID instead of ensembl ID in input file for calculation.
            species: Species for annotation. Only used for cellmarker database. ('Human',['Mouse'])
            weight: Weight threshold for marker filtering from cellranger v1.0 results. (100)
            tissue: Tissue for annotation. you can use `get_model_tissue` to see the available tissues. ('All')
            target: Target to annotation class in Database. (cellmarker,[cancersea,panglaodb])
            celltype: Cell type for annotation. (normal,[cancer])
            norefdb: Only using user-defined marker database for annotation.
            noprint: Do not print any detail results.
            list_tissue: List all available tissues in the database.
            cellrange: Cell sub_type for annotation. (if you input T cell, it will only provide T helper cell, T cytotoxic cell, T regulatory cell, etc.)
        
        """

        #create temp directory
        try:
            if not os.path.isdir('temp'):
                print("...Creating directory {}".format('temp'))
                os.makedirs('temp', exist_ok=True)
        except OSError as e:
            print("...Unable to create directory {}. Reason {}".format('temp',e))

        self.adata=adata
        self.foldchange=foldchange
        self.pvalue=pvalue
        self.output=output
        self.outfmt=outfmt
        self.Gensymbol=Gensymbol
        self.species=species
        self.weight=weight
        self.tissue=tissue
        self.celltype=celltype
        self.norefdb=norefdb
        self.noprint=noprint
        self.list_tissue=list_tissue
        self.target=target
        self.cellrange=cellrange
        if model_path =='':
            self.model_path=data_downloader(url='https://figshare.com/ndownloader/files/41369037',
                                            path='temp/pySCSA_2023_v2_plus.db',title='whole')
        else:
            self.model_path=model_path

    def get_model_tissue(self,species:str="Human")->None:
        r"""List all available tissues in the database.
        
        Arguments:
            species: Species for annotation. Only used for cellmarker database. ('Human',['Mouse'])

        """
        
        anno = Annotator(foldchange=self.foldchange,
                    weight=self.weight,
                    pvalue=self.pvalue,
                    tissue=self.tissue,
                    species=self.species,
                    target=self.target,
                    norefdb=self.norefdb,
                    MarkerDB=None,
                    db=self.model_path,
                    noprint=self.noprint,
                    input="temp/rna.csv",
                    output=self.output,
                    source="scanpy",
                    cluster='all',
                    fc=self.foldchange,
                    outfmt=self.outfmt,
                    celltype=self.celltype,
                    Gensymbol=self.Gensymbol,
                    list_tissue=self.list_tissue,
                    cellrange=self.cellrange)
        anno.load_pickle_module(self.model_path)
        anno.get_list_tissue(species)


    def cell_anno(self,clustertype:str='leiden',
                  cluster:str='all',rank_rep=False)->pd.DataFrame:
        r"""Annotate cell type for each cluster.
        
        Arguments:
            clustertype: Clustering name used in scanpy. (leiden)
            cluster: Only deal with one cluster of marker genes. (all,[1],[1,2,3],[...])
        """

        dat=data_preprocess(self.adata,clustertype=clustertype,path='temp/rna.csv',rank_rep=rank_rep)
        dat.to_csv('temp/rna.csv')

        print('...Auto annotate cell')
        
        p = Process()
        p.run_cmd_p(foldchange=self.foldchange,
                    weight=self.weight,
                    pvalue=self.pvalue,
                    tissue=self.tissue,
                    species=self.species,
                    target=self.target,
                    norefdb=self.norefdb,
                    MarkerDB=None,
                    db=self.model_path,
                    noprint=self.noprint,
                    input="temp/rna.csv",
                    output=self.output,
                    source="scanpy",
                    cluster=cluster,
                    fc=self.foldchange,
                    outfmt=self.outfmt,
                    celltype=self.celltype,
                    Gensymbol=self.Gensymbol,
                    list_tissue=self.list_tissue,
                    cellrange=self.cellrange)
        

        result=pd.read_csv('temp/rna_anno.txt',sep='\t')
        self.result=result
        return result
    
    def cell_anno_print(self)->None:
        r"""print the annotation result
        
        """
        for i in set(self.result['Cluster']):
            test=self.result.loc[self.result['Cluster']==i].iloc[:2]
            if test.iloc[0]['Z-score']>test.iloc[1]['Z-score']*2:
                print('Nice:Cluster:{}\tCell_type:{}\tZ-score:{}'.format(i,test.iloc[0]['Cell Type'],
                                                            np.around(test.iloc[0]['Z-score'],3)))
            else:
                print('Cluster:{}\tCell_type:{}\tZ-score:{}'.format(i,('|').join(test['Cell Type'].values.tolist()),
                                                            ('|').join(np.around(test['Z-score'].values,3).astype(str).tolist())))

    def cell_auto_anno(self,adata:anndata.AnnData,
                       clustertype:str='leiden',key='scsa_celltype')->None:
        r"""Add cell type annotation to anndata.obs['scsa_celltype']
        
        Arguments:
            adata: anndata object
            clustertype: Clustering name used in scanpy. (leiden)
        """
        test_li=[]
        for i in adata.obs[clustertype].value_counts().index:
            if int(i) in self.result['Cluster'].values:
                test_li.append(self.result.loc[self.result['Cluster']==int(i)].iloc[0]['Cell Type'])
            else:
                test_li.append('Unknown')
        scsa_anno=dict(zip([str(i) for i in adata.obs[clustertype].value_counts().index],
            test_li))
        adata.obs[key] = adata.obs[clustertype].map(scsa_anno).astype('category')
        print('...cell type added to {} on obs of anndata'.format(key))

    def get_celltype_marker(self,adata:anndata.AnnData,
                            clustertype:str='leiden',
                            log2fc_min:int=2,scores_type='scores',
                            pval_cutoff:float=0.05,rank:bool=True)->dict:
        r"""Get marker genes for each clusters.
        
        Arguments:
            adata: anndata object
            clustertype: Clustering name used in scanpy. (leiden)
            log2fc_min: Minimum log2 fold change of marker genes. (2)
            pval_cutoff: Maximum p value of marker genes. (0.05)
            rank: Whether to rank genes by wilcoxon test. (True)
            scores_type: The type of scores. can be selected from `scores` and `logfoldchanges`

        Returns:
            cellmarker: A dictionary of marker genes for each clusters.
        """
        print('...get cell type marker')
        cell_marker_dict=get_celltype_marker(adata=adata,
                            clustertype=clustertype,
                            log2fc_min=log2fc_min,scores_type=scores_type,
                            pval_cutoff=pval_cutoff,rank=rank)

        return cell_marker_dict
    


class MetaTiME(object):
    """
    MetaTiME: Meta-components in Tumor immune MicroEnvironment

    Github: https://github.com/yi-zhang/MetaTiME/
    
    """
    
    def __init__(self,adata:anndata.AnnData,mode:str='table'):
        """
        Initialize MetaTiME model

        Arguments:
            adata: anndata object
            mode: choose from ['mecnamedict', 'table', 'meciddict']
                    load manual assigned name for easy understanding of assigned names
                    from file: MeC_anno_name.tsv under mecDIR.
                    Required columns: `['MeC_id', 'Annotation', 'UseForCellStateAnno']` 
                    Required seperator: tab
                    Annotation column NA will be filtered.
                    If you want to use your own annotation, please follow the format of MeC_anno_name.tsv

        """
        check_metatime()
        global metatime_install
        if metatime_install==True:
            global mecs
            global mecmapper
            global annotator
            global config
            from metatime import mecs
            from metatime import mecmapper
            from metatime import annotator
            from metatime import config

        self.adata=adata
        # Load the pre-trained MeCs
        print('...load pre-trained MeCs')
        self.mecmodel = mecs.MetatimeMecs.load_mec_precomputed()

        # Load functional annotation for MetaTiME-TME
        print('...load functional annotation for MetaTiME-TME')
        self.mectable = mecs.load_mecname(mecDIR = config.SCMECDIR, mode =mode )
        self.mecnamedict = mecs.getmecnamedict_ct(self.mectable) 
        
        
    
    def overcluster(self,resolution : float=8, 
                random_state: int= 0, 
                clustercol :str = 'overcluster'):
        """
        Overcluster single cell data to get cluster level cell state annotation

        Arguments:
            resolution: resolution for leiden clustering
            random_state: random state for leiden clustering
            clustercol: column name for cluster level cell state annotation
        
        """

        print('...overclustering using leiden')
        sc.tl.leiden(self.adata, resolution=resolution, key_added = clustercol, random_state=random_state)
        self.clustercol=clustercol
        #self.adata = annotator.overcluster(self.adata,resolution,random_state,clustercol) # this generates a 'overcluster' columns in adata.obs
        
    def predictTiME(self,save_obs_name:str='MetaTiME'):
        """
        Predict TiME celtype for each cell

        Arguments:
            save_obs_name: column name for cell type annotation in adata.obs
        
        """
        print('...projecting MeC scores')
        self.pdata=mecmapper.projectMecAnn(self.adata, self.mecmodel.mec_score)
        projmat, mecscores = annotator.pdataToTable(self.pdata, self.mectable, gcol = self.clustercol)
        projmat, gpred, gpreddict = annotator.annotator(projmat,  self.mecnamedict, gcol = self.clustercol)
        self.adata = annotator.saveToAdata( self.adata, projmat)
        #self.pdata = annotator.saveToPdata( self.pdata, self.adata, projmat )
        self.adata.obs[save_obs_name] = self.adata.obs['{}_{}'.format(save_obs_name,self.clustercol)].str.split(': ').str.get(1)
        #self.pdata.obs[save_obs_name] = self.pdata.obs['{}_{}'.format(save_obs_name,self.clustercol)].str.split(': ').str.get(1)
        self.adata.obs['Major_{}'.format(save_obs_name)]=[i.split('_')[0] for i in self.adata.obs[save_obs_name]]
        #self.pdata.obs['Major_{}'.format(save_obs_name)]=[i.split('_')[0] for i in self.pdata.obs[save_obs_name]]
        print('......The predicted celltype have been saved in obs.{}'.format(save_obs_name))
        print('......The predicted major celltype have been saved in obs.Major_{}'.format(save_obs_name))
        return self.adata
    
    def plot(self,basis:str='X_umap',cluster_key:str='MetaTiME',fontsize:int=8, 
             min_cell:int=5, title=None,figsize:tuple=(6,6),
             dpi:int=80,frameon:bool=False,legend_loc=None,palette=None):
        """
        Plot annotated cells with  non-overlapping fonts.

        Arguments:
            basis: basis for plotting
            cluster_key: column name for cell type annotation in adata.obs
            fontsize: fontsize for plotting
            min_cell: minimum number of cells for plotting
            title: title for plotting
            figsize: figure size for plotting
            dpi: dpi for plotting
            frameon: frameon for plotting
            legend_loc: legend_loc for plotting
            palette: palette for plotting

        Returns:
            fig: figure object
            ax: axis object
        
        """
        import matplotlib.pyplot as plt
        if not title:
            title = cluster_key

        if( min_cell >0 ):
            groupcounts = self.adata.obs.groupby( cluster_key ).count()
            groupcounts = groupcounts[groupcounts.columns[0]]
            group_with_good_counts = groupcounts[groupcounts>= min_cell ].index.tolist()
            self.adata = self.adata[ self.adata.obs[ cluster_key ].isin( group_with_good_counts ) ]

        if palette ==None:
            palette=plt.cycler("color",plt.cm.tab20(np.linspace(0,1,20)))

        with plt.rc_context({"figure.figsize": figsize, "figure.dpi": dpi, "figure.frameon": frameon}):
            #ax = sc.pl.umap(pdata, color="MetaTiME_overcluster", show=False, legend_loc=None, frameon=False, size=30)
            ax = sc.pl.embedding(self.adata, basis=basis,color= cluster_key , show=False, legend_loc=None, add_outline=False, 
                    #legend_loc='on data',legend_fontsize=6, legend_fontoutline=2,
                    title= title, 
                    palette=palette, 
                    #palette=plt.cycler("color",plt.cm.Set1(np.linspace(0,1,9))), 
                    frameon=frameon
                    )
            gen_mpl_labels(
                self.adata,
                cluster_key,
                exclude=("None",),  
                basis=basis,
                ax=ax,
                adjust_kwargs=dict(arrowprops=dict(arrowstyle='-', color='black')),
                text_kwargs=dict(fontsize= fontsize ,weight='bold'),
            )
            fig = ax.get_figure()
            fig.tight_layout()
            #plt.show()
            return( fig,ax )

        


"""
Module to build a bioFAM model
"""

from sys import path
from time import time,sleep
import numpy as np
import scipy as s
#from joblib import Parallel, delayed

from .init_model import initModel
from .utils import *

class buildModel(object):
    def __init__(self, data, dimensionalities, data_opts, model_opts, train_opts):
        self.data = data
        self.dim = dimensionalities
        self.data_opts = data_opts
        self.model_opts = model_opts
        self.train_opts = train_opts

    def createMarkovBlankets(self):
        """ Define the markov blankets """
        pass

    def createSchedule(self):
        """ Define the schedule of updates """
        pass

    def build_nodes(self):
        """ Build all nodes """
        pass

    def get_nodes(self):
        """ Get all nodes """
        return self.init_model.getNodes()




class buildBiofam(buildModel):
    def  __init__(self, *args, **kwargs): 
        super(buildBiofam, self).__init__(*args, **kwargs)

    def main(self):
        # create an instance of initModel
        self.init_model = initModel(
            dim = self.dim, 
            data = self.data, 
            lik = self.model_opts["likelihoods"], 
            groups = self.data_opts['samples_groups'],
            seed = self.train_opts['seed']
        )

        # Build all nodes
        self.build_nodes()

        # Define markov blankets
        self.createMarkovBlankets()

    def build_nodes(self):
        """ Method to build all nodes """

        # Build general nodes
        self.build_Z()
        self.build_W()
        self.build_Tau()
        self.build_Y() # important to keep Y last. NAs get replaced by 0 and can mislead PCA initialisation

        # define ARD sparsity per sample group (on Z)
        if self.model_opts['ard_factors']:
            self.build_AlphaZ()

        # define ARD sparsity per feature_group, or view (on W)
        if self.model_opts['ard_weights']:
            self.build_AlphaW()

        # define feature-wise spike and slab sparsity in Z
        if self.model_opts['spikeslab_factors']:
            self.build_ThetaZ()

        # define feature-wise spike and slab sparsity in W
        if self.model_opts['spikeslab_weights']:
            self.build_ThetaW()

    def build_Y(self):
        """ Build node Y for the observations """
        self.init_model.initY()

    def build_Z(self):
        """ Build node Z for the factors or latent variables """
        if self.model_opts['spikeslab_factors']:
            # self.init_model.initSZ(qmean_T1=0)
            # self.init_model.initSZ(qmean_T1="random")
            self.init_model.initSZ(qmean_T1="pca", Y=self.data, impute=True, weight_views = self.train_opts['weight_views'])
        else:
            # self.init_model.initZ(qmean=0)
            # self.init_model.initZ(qmean="random")
            self.init_model.initZ(
                qmean="pca", 
                Y=self.data, 
                impute=True,
                weight_views = self.train_opts['weight_views']
            )

    def build_W(self):
        """ Build node W for the weights """
        if self.model_opts['spikeslab_weights']:
            self.init_model.initSW(qmean_S1=0)
            # self.init_model.initSW(qmean_S1="random", Y=self.data)
            # self.init_model.initSW(qmean_S1="pca", Y=self.data)
        else:
            self.init_model.initW(qmean=0)
            # self.init_model.initW(qmean="random", Y=self.data)
            # self.init_model.initW(qmean="pca", Y=self.data)

    def build_Tau(self):
        # TODO sort out how to choose where to use Tau
        # initTau_qE = 100.
        initTau_qE = None
            
        self.init_model.initTau(qE=initTau_qE)

    def build_AlphaZ(self):
        """ Build node AlphaZ for the ARD prior on the factors """

        # ARD prior per sample group
        self.init_model.initAlphaZ()

    def build_AlphaW(self):
        """ Build node AlphaW for the ARD prior on the weights"""

        # ARD prior per factor and feature_group (view)
        self.init_model.initAlphaW()

    def build_ThetaZ(self):
        # TODO use mixed theta node instead when fixed in update and init -> should be ok then for intercept
        """ Build node ThetaZ for the Spike and Slab prior on the factors """

        # Initialise hyperparameters for the ThetaZ prior
        initTheta_a = 1.
        initTheta_b = 1.
        initTheta_qE = None
        # initTheta_qE = 1.

        self.init_model.initThetaZ(self.data_opts['samples_groups'], qa=initTheta_a, qb=initTheta_b, qE=initTheta_qE)

    def build_ThetaW(self):
        """ Build node ThetaW for the Spike and Slab prior on the weights """

        # Initialise hyperparameters for the ThetaW prior
        initTheta_a = 1.
        initTheta_b = 1.
        initTheta_qE = None
        # initTheta_qE = 1.

        self.init_model.initThetaW(qa=initTheta_a, qb=initTheta_b, qE=initTheta_qE)

    def createMarkovBlankets(self):
        """ Define the markov blankets """

        # Fetch all nodes
        nodes = self.get_nodes()

        # Define basic connections
        nodes['Y'].addMarkovBlanket(Z=nodes["Z"], W=nodes["W"], Tau=nodes["Tau"])
        nodes['Z'].addMarkovBlanket(Y=nodes["Y"], W=nodes["W"], Tau=nodes["Tau"])
        nodes['W'].addMarkovBlanket(Y=nodes["Y"], Z=nodes["Z"], Tau=nodes["Tau"])
        nodes['Tau'].addMarkovBlanket(Y=nodes["Y"], W=nodes["W"], Z=nodes["Z"])

        # Add ThetaZ in the markov blanket of Z and viceversa if Spike and Slab prior on Z
        if self.model_opts['spikeslab_factors']:
            nodes['Z'].addMarkovBlanket(ThetaZ=nodes["ThetaZ"])
            nodes["ThetaZ"].addMarkovBlanket(Z=nodes["Z"])

        # Add ThetaW in the markov blanket of W and viceversa if Spike and Slab prior on W
        if self.model_opts['spikeslab_weights']:
            nodes['W'].addMarkovBlanket(ThetaW=nodes["ThetaW"])
            nodes["ThetaW"].addMarkovBlanket(W=nodes["W"])

        # Add AlphaZ in the markov blanket of Z and viceversa if ARD prior on Z
        if self.model_opts['ard_factors']:
                nodes['AlphaZ'].addMarkovBlanket(Z=nodes['Z'])
                nodes['Z'].addMarkovBlanket(AlphaZ=nodes['AlphaZ'])

        # Add AlphaW in the markov blanket of W and viceversa if ARD prior on W
        if self.model_opts['ard_weights']:
            nodes['AlphaW'].addMarkovBlanket(W=nodes['W'])
            nodes['W'].addMarkovBlanket(AlphaW=nodes['AlphaW'])


class build_mofa_smooth(buildBiofam):
    def __init__(self, data, sample_cov, dimensionalities, data_opts, model_opts, train_opts, smooth_opts):
        super(build_mofa_smooth, self).__init__(data, dimensionalities, data_opts, model_opts, train_opts)
        self.sample_cov = sample_cov
        self.smooth_opts = smooth_opts

    def main(self):
        # create an instance of initModel
        self.init_model = initModel(
            dim = self.dim, 
            data = self.data, 
            lik = self.model_opts["likelihoods"], 
            groups = self.data_opts['samples_groups'],
            seed = self.train_opts['seed']
        )

        # Build all nodes
        self.build_nodes()

        # Define markov blankets
        self.createMarkovBlankets()

    def build_Z(self):
        """ Build node Z for the factors """
        self.init_model.initZ_smooth(
            qmean="pca", 
            Y=self.data, 
            impute=True,
            weight_views = self.train_opts['weight_views']
        )

    def build_ZgU(self):
        """ Build node for Z given U for the factors or latent variables conditioned on inducing points"""

        self.init_model.initZgU(
            qmean="pca", 
            Y=self.data, 
            impute=True,
            idx_inducing = self.smooth_opts['idx_inducing'],
            weight_views = self.train_opts['weight_views']
        )

    def build_U(self):
        """ Build node U for the inducing points of latent variable GP """

        # initialise U by PCA (no use of GP prior)
        self.init_model.initU(
            idx_inducing = self.smooth_opts['idx_inducing'],
            weight_views = self.train_opts['weight_views']
        )

    def build_Sigma(self):
        """ Build node Sigma for the GP prior on the factors """

        # TO-DO: assert on build that G >1 if warping
        # TO-DO: exclude idx_inducing when using warping

        # Sparse GPs
        if self.smooth_opts['sparseGP'] is True:
            # Warping
            if self.smooth_opts['warping'] is True:
                raise NotImplementedError
            # Non-Warping
            else:
                self.init_model.initSigma_sparse(
                    self.sample_cov,
                    start_opt = self.smooth_opts['start_opt'],
                    n_grid = self.smooth_opts['n_grid'],
                    idx_inducing = self.smooth_opts['idx_inducing'],
                    opt_freq = self.smooth_opts['opt_freq'],
                    model_groups = self.smooth_opts['model_groups']
                )

        # Non-sparse GPs
        else:
            # Warping
            if self.smooth_opts['warping'] is True:
                self.init_model.initSigma_warping(
                    sample_cov = self.sample_cov,
                    start_opt = self.smooth_opts['start_opt'],
                    n_grid = self.smooth_opts['n_grid'],
                    warping_freq = self.smooth_opts['warping_freq'],
                    warping_ref = self.smooth_opts['warping_ref'],
                    warping_open_begin = self.smooth_opts['warping_open_begin'],
                    warping_open_end = self.smooth_opts['warping_open_end'],
                    warping_groups = self.smooth_opts['warping_groups'],
                    opt_freq = self.smooth_opts['opt_freq'],
                    model_groups = self.smooth_opts['model_groups']#,
                    # use_gpytorch  = self.model_opts['use_gpytorch']
                )
            # Non-warping
            else:
                self.init_model.initSigma(
                    self.sample_cov,
                    start_opt = self.smooth_opts['start_opt'],
                    n_grid = self.smooth_opts['n_grid'],
                    # idx_inducing = self.smooth_opts['idx_inducing'],
                    # warping = self.smooth_opts['warping'],
                    # warping_freq = self.smooth_opts['warping_freq'],
                    # warping_ref = self.smooth_opts['warping_ref'],
                    # warping_open_begin = self.smooth_opts['warping_open_begin'],
                    # warping_open_end = self.smooth_opts['warping_open_end'],
                    opt_freq = self.smooth_opts['opt_freq'],
                    model_groups = self.smooth_opts['model_groups']#,
                    # use_gpytorch  = self.model_opts['use_gpytorch']
                )

    def build_nodes(self):
        """ Method to build all nodes """

        # smooth factors nodes
        if self.smooth_opts["sparseGP"]:
            self.build_U()
            self.build_ZgU()
        else:
            self.build_Z()

        # Build general nodes
        self.build_W()
        self.build_Tau()
        self.build_Y() # important to keep Y last. NAs get replaced by 0 and can mislead PCA initialisation

        # define ARD sparsity per feature_group, or view (on W)
        if self.model_opts['ard_weights']:
            self.build_AlphaW()

        # define Gaussian process prior on Z
        self.build_Sigma()

        # define feature-wise spike and slab sparsity in W
        if self.model_opts['spikeslab_weights']:
            self.build_ThetaW()

    def createMarkovBlankets(self):
        """ Define the markov blankets """

        # Fetch all nodes
        nodes = self.get_nodes()

        # Define basic connections
        nodes['Y'].addMarkovBlanket(Z=nodes["Z"], W=nodes["W"], Tau=nodes["Tau"])
        nodes['Z'].addMarkovBlanket(Y=nodes["Y"], W=nodes["W"], Tau=nodes["Tau"])
        nodes['W'].addMarkovBlanket(Y=nodes["Y"], Z=nodes["Z"], Tau=nodes["Tau"])
        nodes['Tau'].addMarkovBlanket(Y=nodes["Y"], W=nodes["W"], Z=nodes["Z"])

        
        # Add Sigma in the markov blanket of Z and viceversa if GP prior on Z
        if self.smooth_opts['sparseGP']:
            nodes['Z'].addMarkovBlanket(U=nodes["U"], Sigma = nodes['Sigma'])
            nodes["Sigma"].addMarkovBlanket(U=nodes["U"])
            nodes['U'].addMarkovBlanket(Z=nodes["Z"], Sigma = nodes['Sigma'], Y=nodes["Y"], W=nodes["W"], Tau=nodes["Tau"])
        else:
            nodes['Z'].addMarkovBlanket(Sigma=nodes["Sigma"])
            nodes["Sigma"].addMarkovBlanket(Z=nodes["Z"])

        # Add ThetaW in the markov blanket of W and viceversa if Spike and Slab prior on W
        if self.model_opts['spikeslab_weights']:
            nodes['W'].addMarkovBlanket(ThetaW=nodes["ThetaW"])
            nodes["ThetaW"].addMarkovBlanket(W=nodes["W"])


        # Add AlphaW in the markov blanket of W and viceversa if ARD prior on W
        if self.model_opts['ard_weights']:
            nodes['AlphaW'].addMarkovBlanket(W=nodes['W'])
            nodes['W'].addMarkovBlanket(AlphaW=nodes['AlphaW'])



from __future__ import division

import scipy as s
import numpy as np

from .variational_nodes import *
from ..gp_utils import *
from .. import gpu_utils

class Kc_Node(Node):
    """
    Sigma node to model the covariance structure K_c along the given covariate(s)
    This node constructs the covariate kernel matrix on a grid of lengthscale parameters and
    keeps track of the ELBO-optimal idx.

    By default a squared exponential kernel is used:
    KC_{ij} = SE(i,j) = exp(-0.5 * (i-j)^2 / (l**2))

    PARAMETERS
    ----------
    dim: dimensionality of the node (= number of latent factors x number of covariate observations)
    covariates: unique covariates for construction of the covariate part of the covariance matrix
    n_grid: number of grid points for the lengthscale parameter
    """

    def __init__(self, dim, covariates, n_grid=10, cov4grid = None, spectral_decomp = True):
        super().__init__(dim)
        self.covariates = covariates                                # covariate
        self.C = dim[1]                                             # number of observations for covariate
        self.K = dim[0]                                             # number of latent processes
        self.n_grid = n_grid                                        # number of grid points to optimize lengthscale on
        if cov4grid is None:
            self.cov4grid = covariates                              # covariates use to build grid - for consistsney with non-spasre GP
        else:
            self.cov4grid = cov4grid                                    #(can differ from covariates, e.g. if kernel is built on subset on inducing points)
        self.gridix = np.zeros(self.K, dtype = np.int8)             # index of the grid values for lengthscale selected per factor


        # initialize components in node
        self.compute4init(spectral_decomp)

    def compute4init(self, spectral_decomp = True):
        """
        Function to initiailize the grid, kernel matrix and spectral decomposition
        """

        # use all covariate to determine grid
        self.l_grid = get_l_grid(self.cov4grid, n_grid = self.n_grid)

        # add the diagonal covariance (lengthscale 0) to the grid
        self.l_grid = np.insert(self.l_grid, 0, 0)
        self.n_grid += 1

        # initialise kernel matrix
        if not spectral_decomp:
            self.Kmat = np.zeros([self.n_grid, self.C, self.C])  # kernel matrix on lengthscale grid

        # initialise spectral decomposition
        else:
            self.V = np.zeros([self.n_grid, self.C, self.C])    # eigenvectors of kernel matrix on lengthscale grid
            self.D = np.zeros([self.n_grid, self.C])            # eigenvalues of kernel matrix on lengthscale grid

        # compute for each lengthscale the kernel matrix
        self.compute_kernel(spectral_decomp = spectral_decomp)

    def compute_kernel(self, spectral_decomp = True):
        """
        Function to compute kernel matrix for all lengthscales
        """

        for i in range(self.n_grid):
            self.compute_kernel_at_gridpoint(i, spectral_decomp = spectral_decomp)


    def compute_kernel_at_gridpoint(self, i, spectral_decomp = True):

        # build kernel matrix based on given covariance function
        if not spectral_decomp:
            self.Kmat[i, :, :] = SE(self.covariates, self.l_grid[i], zeta=0)
            # self.Kmat[i, :, :] = Cauchy(self.sample_cov_transformed, self.l_grid[i], zeta=0)
        else:
            Kmat = SE(self.covariates, self.l_grid[i], zeta=0)
            # compute spectral decomposition
            # Kc = VDV^T with V^T V = I
            self.D[i, :], self.V[i, :, :] = s.linalg.eigh(Kmat)


    def get_ls(self):
        """
        Method to fetch ELBO-optimal length-scales
        """
        ls = np.array([self.l_grid[i] for i in self.gridix])
        return ls

    def get_kernel_components_k(self, k):
        """
        Method to ELBO optimal components of kernel for given factor k
        """
        best_ls_idx = self.get_best_lidx(k)
        return self.V[best_ls_idx, :, :], self.D[best_ls_idx,:]

    def get_best_lidx(self, k):
        return self.gridix[k]

    def removeFactors(self, idx, axis=1):
        self.gridix = np.delete(self.gridix, axis=0, obj=idx)
        self.updateDim(0, self.dim[0] - len(idx))
        self.K = self.K - 1


    def set_gridix(self, lidx, k):
        self.gridix[k] = lidx
        # no recomputation required as stored on grid

    def eval_at_newpoints_k(self, new_cov, k):

        Kc_new = SE(new_cov, self.l_grid[self.gridix[k]], zeta=0)

        return Kc_new



import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
from sklearn import linear_model
from scipy.stats import norm
from scipy import stats
import networkx as nx
import seaborn as sns
import pandas as pd
from scipy.cluster import hierarchy  
from scipy import cluster   
from sklearn import decomposition as skldec 
from gseapy.plot import gseaplot,GSEAPlot


from scipy.spatial.distance import pdist
from scipy.cluster.hierarchy import linkage,dendrogram


import gseapy as gp
from gseapy.plot import barplot, dotplot
from ..utils import plot_text_set
import matplotlib

"""
def enrichment_KEGG(gene_list,
                    gene_sets=['KEGG_2019_Human'],
                    organism='Human',
                    description='test_name',
                    outdir='enrichment_kegg',
                    cutoff=0.5):

    '''
    Gene enrichment analysis of KEGG

    Parameters
    ----------
    gene_list:list
        The gene set to be enrichment analyzed
    gene_sets:list
        The gene_set of enrichr library
        Input Enrichr Libraries (https://maayanlab.cloud/Enrichr/#stats)
    organism:str
        Select from (human, mouse, yeast, fly, fish, worm)
    description:str
        The title of enrichment
    outdir:str
        The savedir of enrichment
    cutoff:float
        Show enriched terms which Adjusted P-value < cutoff.

    Returns
    ----------
    res:pandas.DataFrame
        stores your last query
    '''

    enr = gp.enrichr(gene_list=gene_list,
                 gene_sets=gene_sets,
                 organism=organism, # don't forget to set organism to the one you desired! e.g. Yeast
                 description=description,
                 outdir=outdir,
                 # no_plot=True,
                 cutoff=cutoff # test dataset, use lower value from range(0,1)
                )
    subp=dotplot(enr.res2d, title=description,cmap='seismic')
    print(subp)
    return enr.res2d

def enrichment_GO(gene_list,
                    go_mode='Bio',
                    organism='Human',
                    description='test_name',
                    outdir='enrichment_go',
                    cutoff=0.5):

    '''
    Gene enrichment analysis of GO

    Parameters
    ----------
    gene_list:list
        The gene set to be enrichment analyzed
    go_mode:str
        The module of GO include:'Bio','Cell','Mole'
    organism:str
        Select from (human, mouse, yeast, fly, fish, worm)
    description:str
        The title of enrichment
    outdir:str
        The savedir of enrichment
    cutoff:float
        Show enriched terms which Adjusted P-value < cutoff.

    Returns
    ----------
    result:pandas.DataFrame
        stores your last query
    '''
    if(go_mode=='Bio'):
        geneset='GO_Biological_Process_2018'
    if(go_mode=='Cell'):
        geneset='GO_Cellular_Component_2018'
    if(go_mode=='Mole'):
        geneset='GO_Molecular_Function_2018'
    enr = gp.enrichr(gene_list=gene_list,
                 gene_sets=geneset,
                 organism=organism, # don't forget to set organism to the one you desired! e.g. Yeast
                 description=description,
                 outdir=outdir,
                 # no_plot=True,
                 cutoff=cutoff # test dataset, use lower value from range(0,1)
                )
    subp=dotplot(enr.res2d, title=description,cmap='seismic')
    print(subp)
    return enr.res2d

def enrichment_GSEA(data,
                   gene_sets='KEGG_2016',
                   processes=4,
                   permutation_num=100,
                   outdir='prerank_report_kegg',
                   seed=6):
    '''
    Gene enrichment analysis of GSEA

    Parameters
    ----------
    data:pandas.DataFrame
        The result of Find_DEG(function in DeGene.py)
    gene_sets:list
        The gene_set of enrichr library
        Input Enrichr Libraries (https://maayanlab.cloud/Enrichr/#stats)
    processes:int
        CPU number
    permutation_num:int
        Number of permutations for significance computation. Default: 1000.
    outdir:str
        The savedir of enrichment
    seed:int
        Random seed

    Returns
    ----------
    result:Return a Prerank obj. 
    All results store to  a dictionary, obj.results,
         where contains::

             | {es: enrichment score,
             |  nes: normalized enrichment score,
             |  p: P-value,
             |  fdr: FDR,
             |  size: gene set size,
             |  matched_size: genes matched to the data,
             |  genes: gene names from the data set
             |  ledge_genes: leading edge genes}
    '''


    rnk=pd.DataFrame(columns=['genename','FoldChange'])
    rnk['genename']=data.index
    rnk['FoldChange']=data['FoldChange'].tolist()
    rnk1=rnk.drop_duplicates(['genename'])
    rnk1=rnk1.sort_values(by='FoldChange', ascending=False)
    
    pre_res = gp.prerank(rnk=rnk1, gene_sets=gene_sets,
                     processes=processes,
                     permutation_num=permutation_num, # reduce number to speed up testing
                     outdir=outdir, format='png', seed=seed)
    pre_res.res2d.sort_index().to_csv('GSEA_result.csv')
    return pre_res

def Plot_GSEA(data,num=0):

    '''
    Plot the GSEA result figure

    Parameters
    ----------
    data:prerank obj
        The result of enrichment_GSEA
    num:int
        The sequence of pathway drawn 
        Default:0(the first pathway)

    '''
    terms = data.res2d.index
    from gseapy.plot import gseaplot
    # to save your figure, make sure that ofname is not None
    gseaplot(rank_metric=data.ranking, term=terms[num], **data.results[terms[num]])
"""
def geneset_enrichment(gene_list:list,pathways_dict:dict,
                       pvalue_threshold:float=0.05,pvalue_type:str='auto',
                       organism:str='Human',description:str='None',
                       background:list=None,
                       outdir:str='./enrichr',cutoff:float=0.5)->pd.DataFrame:
    """
    Performs gene set enrichment analysis using Enrichr API.

    Arguments:
        gene_list: List of gene symbols to be tested for enrichment.
        pathways_dict: Dictionary of pathway library names and corresponding Enrichr API URLs.
        pvalue_threshold: P-value threshold for significant pathways. Default is 0.05.
        pvalue_type: Type of p-value correction to use. 'auto' uses Benjamini-Hochberg correction,for small gene sets (<500 genes) and Bonferroni correction for larger gene sets.,'bh' uses only Benjamini-Hochberg correction. 'bonferroni' uses only Bonferroni correction.,Default is 'auto'.
        organism: Organism of the input gene list. Default is 'Human'.
        description: Description of the input gene list. Default is 'None'.
        background: Background gene list to use for enrichment analysis. Default is None. If None, the background gene list is automatically set to the organism-specific gene list.
        outdir: Output directory for Enrichr results. Default is './enrichr'.
        cutoff: Show enriched terms which Adjusted P-value < cutoff. Default is 0.5.

    Returns:
        enrich_res: A pandas DataFrame containing the enrichment results.


    """
    if background is None:
        if (organism == 'Mouse') or (organism == 'mouse') or (organism == 'mm'):
            background='mmusculus_gene_ensembl'
        elif (organism == 'Human') or (organism == 'human') or (organism == 'hs'):
            background='hsapiens_gene_ensembl'

    enr = gp.enrichr(gene_list=gene_list,
                 gene_sets=pathways_dict,
                 organism=organism, # don't forget to set organism to the one you desired! e.g. Yeast
                 description=description,
                 background=background,
                 outdir=outdir,
                 cutoff=cutoff # test dataset, use lower value from range(0,1)
                )
    if pvalue_type=='auto':
        if enr.res2d.shape[0]>100:
            enrich_res=enr.res2d[enr.res2d['Adjusted P-value']<pvalue_threshold]
            enrich_res['logp']=-np.log(enrich_res['Adjusted P-value'])
        else:
            enrich_res=enr.res2d[enr.res2d['P-value']<pvalue_threshold]
            enrich_res['logp']=-np.log(enrich_res['P-value'])
    elif pvalue_type=='adjust':
        enrich_res=enr.res2d[enr.res2d['Adjusted P-value']<pvalue_threshold]
        enrich_res['logp']=-np.log(enrich_res['Adjusted P-value'])
    else:
        enrich_res=enr.res2d[enr.res2d['P-value']<pvalue_threshold]
        enrich_res['logp']=-np.log(enrich_res['P-value'])
    enrich_res['logc']=np.log(enrich_res['Odds Ratio'])
    enrich_res['num']=[int(i.split('/')[0]) for i in enrich_res['Overlap']]
    enrich_res['fraction']=[int(i.split('/')[0])/int(i.split('/')[1]) for i in enrich_res['Overlap']]
    return enrich_res



def geneset_enrichment_GSEA(gene_rnk:pd.DataFrame,pathways_dict:dict,
                            processes:int=8,
                     permutation_num:int=100, # reduce number to speed up testing
                     outdir:str='./enrichr_gsea', format:str='png', seed:int=112)->dict:
    """
    Enrichment analysis using GSEA

    Arguments:
        gene_rnk: pre-ranked correlation table or pandas DataFrame. Same input with ``GSEA`` .rnk file.
        pathways_dict: Dictionary of pathway library names and corresponding Enrichr API URLs.
        processes: Number of Processes you are going to use. Default: 8.
        permutation_num: Number of permutations for significance computation. Default: 100.
        outdir: Output directory for Enrichr results. Default is './enrichr_gsea'.
        format: Matplotlib figure format. Default: 'png'.
        seed: Random seed. Default: 112.

    Returns:
        pre_res: A prerank object containing the enrichment results.
    
    """
    pre_res = gp.prerank(rnk=gene_rnk, gene_sets=pathways_dict,
                     processes=processes,
                     permutation_num=permutation_num, # reduce number to speed up testing
                     outdir=outdir, format=format, seed=seed)
    return pre_res
    enrich_res=pre_res.res2d[pre_res.res2d['fdr']<0.05]
    enrich_res['logp']=-np.log(enrich_res['fdr']+0.0001)
    enrich_res['logc']=enrich_res['nes']
    enrich_res['num']=enrich_res['matched_size']
    enrich_res['fraction']=enrich_res['matched_size']/enrich_res['geneset_size']
    enrich_res['Term']=enrich_res.index.tolist()
    enrich_res['P-value']=enrich_res['fdr']
    return enrich_res


def geneset_plot(enrich_res,num:int=10,node_size:list=[5,10,15],
                        cax_loc:int=2,cax_fontsize:int=12,
                        fig_title:str='',fig_xlabel:str='Fractions of genes',
                        figsize:tuple=(2,4),cmap:str='YlGnBu',
                        text_knock:int=2,text_maxsize:int=20)->matplotlib.axes._axes.Axes:
    """
    Plot the gene set enrichment result.

    Arguments:
        num: The number of enriched terms to plot. Default is 10.
        node_size: A list of integers defining the size of nodes in the plot. Default is [5,10,15].
        cax_loc: The location of the colorbar on the plot. Default is 2.
        cax_fontsize: The fontsize of the colorbar label. Default is 12.
        fig_title: The title of the plot. Default is an empty string.
        fig_xlabel: The label of the x-axis. Default is 'Fractions of genes'.
        figsize: The size of the plot. Default is (2,4).
        cmap: The colormap to use for the plot. Default is 'YlGnBu'.
        text_knock: The number of characters to knock off the end of the term name. Default is 2.
        text_maxsize: The maximum fontsize of the term names. Default is 20.

    Returns:
        A matplotlib.axes.Axes object.
    
    """
    fig, ax = plt.subplots(figsize=figsize)
    plot_data2=enrich_res.sort_values('P-value')[:num].sort_values('logc')
    st=ax.scatter(plot_data2['fraction'],range(len(plot_data2['logc'])),
            s=plot_data2['num']*10,linewidths=1,edgecolors='black',c=plot_data2['logp'],cmap=cmap)
    ax.yaxis.tick_right()
    plt.yticks(range(len(plot_data2['fraction'])),[plot_text_set(i.split('(')[0],text_knock=text_knock,text_maxsize=text_maxsize) for i in plot_data2['Term']],
            fontsize=10,)
    plt.xticks(fontsize=12,)
    plt.title(fig_title,fontsize=12)
    plt.xlabel(fig_xlabel,fontsize=12)

    #fig = plt.gcf()
    cax = fig.add_axes([cax_loc, 0.55, 0.5, 0.02])
    cb=fig.colorbar(st,shrink=0.25,cax=cax,orientation='horizontal')
    cb.set_label(r'$−Log_{10}(P_{adjusted})$',fontdict={'size':cax_fontsize})

    gl_li=[]
    for i in node_size:
        gl_li.append(ax.scatter([],[], s=i*10, marker='o', color='white',edgecolors='black'))

    plt.legend(gl_li,
        [str(i) for i in node_size],
        loc='lower left',
        ncol=3,bbox_to_anchor=(-0.45, -13),
        fontsize=cax_fontsize)
    return ax

class pyGSE(object):

    def __init__(self,gene_list:list,pathways_dict:dict,pvalue_threshold:float=0.05,pvalue_type:str='auto',
                 background=None,organism:str='Human',description:str='None',outdir:str='./enrichr',cutoff:float=0.5) -> None:
        """Initialize the pyGSEA class.

        Arguments:
            gene_list: A list of genes.
            pathways_dict: A dictionary of pathways.
            pvalue_threshold: The p-value threshold for enrichment. Default is 0.05.
            pvalue_type: The p-value type. Default is 'auto'.
            organism: The organism. Default is 'Human'.
            description: The description. Default is 'None'.
            outdir: The output directory. Default is './enrichr'.
            cutoff: The cutoff for enrichment. Default is 0.5.
        
        """

        self.gene_list=gene_list
        self.pathways_dict=pathways_dict
        self.pvalue_threshold=pvalue_threshold
        self.pvalue_type=pvalue_type
        self.organism=organism
        self.description=description
        self.outdir=outdir
        self.cutoff=cutoff
        if background is None:
            if (organism == 'Mouse') or (organism == 'mouse') or (organism == 'mm'):
                background='mmusculus_gene_ensembl'
            elif (organism == 'Human') or (organism == 'human') or (organism == 'hs'):
                background='hsapiens_gene_ensembl'
            self.background=background
        else:
            self.background=background
    
    def enrichment(self):
        """gene set enrichment analysis.
        
        Returns:
            A pandas.DataFrame object containing the enrichment results.
        """

        enrich_res=geneset_enrichment(self.gene_list,self.pathways_dict,self.pvalue_threshold,self.pvalue_type,
                                  self.organism,self.description,self.background,self.outdir,self.cutoff)
        self.enrich_res=enrich_res
        return enrich_res
    
    
    def plot_enrichment(self,num:int=10,node_size:list=[5,10,15],
                        cax_loc:int=2,cax_fontsize:int=12,
                        fig_title:str='',fig_xlabel:str='Fractions of genes',
                        figsize:tuple=(2,4),cmap:str='YlGnBu',text_knock:int=2,text_maxsize:int=20)->matplotlib.axes._axes.Axes:
        
        """Plot the gene set enrichment result.
        
        Arguments:
            num: The number of enriched terms to plot. Default is 10.
            node_size: A list of integers defining the size of nodes in the plot. Default is [5,10,15].
            cax_loc: The location of the colorbar on the plot. Default is 2.
            cax_fontsize: The fontsize of the colorbar label. Default is 12.
            fig_title: The title of the plot. Default is an empty string.
            fig_xlabel: The label of the x-axis. Default is 'Fractions of genes'.
            figsize: The size of the plot. Default is (2,4).
            cmap: The colormap to use for the plot. Default is 'YlGnBu'.

        Returns:
            A matplotlib.axes.Axes object.
        """
        return geneset_plot(self.enrich_res,num,node_size,cax_loc,cax_fontsize,
                            fig_title,fig_xlabel,figsize,cmap,text_knock,text_maxsize)
    
class pyGSEA(object):

    def __init__(self,gene_rnk:pd.DataFrame,pathways_dict:dict,
                 processes:int=8,permutation_num:int=100,
                 outdir:str='./enrichr_gsea',cutoff:float=0.5) -> None:
        """Initialize the pyGSEA class.

        Arguments:
            gene_rnk: pre-ranked correlation table or pandas DataFrame. Same input with ``GSEA`` .rnk file.
            pathways_dict: Dictionary of pathway library names and corresponding Enrichr API URLs.
            processes: Number of Processes you are going to use. Default: 8.
            permutation_num: Number of permutations for significance computation. Default: 100.
            outdir: Output directory for Enrichr results. Default is './enrichr_gsea'.
            cutoff: The cutoff for enrichment. Default is 0.5.
        """

        self.gene_rnk=gene_rnk
        self.pathways_dict=pathways_dict
        self.processes=processes
        self.permutation_num=permutation_num
        self.outdir=outdir
        self.cutoff=cutoff
    
    
    def enrichment(self,format:str='png', seed:int=112)->pd.DataFrame:
        """gene set enrichment analysis.
        
        Arguments:
            format: Matplotlib figure format. Default: 'png'.
            seed: Random seed. Default: 112.
        
        Returns:
            enrich_res:A pandas.DataFrame object containing the enrichment results.
        """

        
        pre_res=geneset_enrichment_GSEA(self.gene_rnk,self.pathways_dict,
                                           self.processes,self.permutation_num,
                                           self.outdir,format,seed)
        self.pre_res=pre_res
        enrich_res=pre_res.res2d[pre_res.res2d['fdr']<0.05]
        enrich_res['logp']=-np.log(enrich_res['fdr']+0.0001)
        enrich_res['logc']=enrich_res['nes']
        enrich_res['num']=enrich_res['matched_size']
        enrich_res['fraction']=enrich_res['matched_size']/enrich_res['geneset_size']
        enrich_res['Term']=enrich_res.index.tolist()
        enrich_res['P-value']=enrich_res['fdr']
        self.enrich_res=enrich_res
        return enrich_res
    
    def plot_gsea(self,term_num:int=0,
                  gene_set_title:str='',
                  figsize:tuple=(3,4),
                  cmap:str='RdBu_r',
                  title_fontsize:int=12,
                  title_y:float=0.95)->matplotlib.figure.Figure:
        """Plot the gene set enrichment result.
        
        Arguments:
            term_num: The number of enriched terms to plot. Default is 0.
            gene_set_title: The title of the plot. Default is an empty string.
            figsize: The size of the plot. Default is (3,4).
            cmap: The colormap to use for the plot. Default is 'RdBu_r'.
            title_fontsize: The fontsize of the title. Default is 12.
            title_y: The y coordinate of the title. Default is 0.95.

        Returns:
            fig: A matplotlib.figure.Figure object.
        """
        
        terms = self.enrich_res.index
        g = GSEAPlot(
        rank_metric=self.pre_res.ranking, term=terms[term_num],figsize=figsize,cmap=cmap,
            **self.pre_res.results[terms[term_num]]
            )
        if gene_set_title=='':
            g.fig.suptitle(terms[term_num],fontsize=title_fontsize,y=title_y)
        else:
            g.fig.suptitle(gene_set_title,fontsize=title_fontsize,y=title_y)
        g.add_axes()
        return g.fig
    
    
    def plot_enrichment(self,num:int=10,node_size:list=[5,10,15],
                        cax_loc:int=2,cax_fontsize:int=12,
                        fig_title:str='',fig_xlabel:str='Fractions of genes',
                        figsize:tuple=(2,4),cmap:str='YlGnBu',
                        text_knock:int=2,text_maxsize:int=20)->matplotlib.axes._axes.Axes:
        
        """Plot the gene set enrichment result.
        
        Arguments:
            num: The number of enriched terms to plot. Default is 10.
            node_size: A list of integers defining the size of nodes in the plot. Default is [5,10,15].
            cax_loc: The location of the colorbar on the plot. Default is 2.
            cax_fontsize: The fontsize of the colorbar label. Default is 12.
            fig_title: The title of the plot. Default is an empty string.
            fig_xlabel: The label of the x-axis. Default is 'Fractions of genes'.
            figsize: The size of the plot. Default is (2,4).
            cmap: The colormap to use for the plot. Default is 'YlGnBu'.
            text_knock: The number of terms to knock out for text labels. Default is 2.
            text_maxsize: The maximum fontsize of text labels. Default is 20.

        Returns:
            ax: A matplotlib.axes.Axes object.
        """
        return geneset_plot(self.enrich_res,num,node_size,cax_loc,cax_fontsize,
                            fig_title,fig_xlabel,figsize,cmap,text_knock,text_maxsize)


import scipy as s
import numpy as np
import scipy.stats as stats
import sys
import scipy.spatial as SS
import math
import seaborn as sns
import matplotlib.pyplot as plt
from ..core.gp_utils import covar_to_corr

def simulate_data(N=200, seed=1234567, views = ["0", "1", "2", "3"], D = [500, 200, 500, 200], noise_level = 1,
                  K = 4, G = 1, lscales = [0.2, 0.8, 0.0, 0.0], sample_cov = "equidistant", scales = [1, 0.8, 0, 0],
                  shared = True, plot = False, alpha = None):
    """
    Function to simulate test data for MOFA (without ARD or spike-and-slab on factors)

    N: Number of time points/ samples per group
    seed: seed to use for simulation
    views: list of view names
    K: number of factors
    G: Number of groups
    D: list of number of features per view (same length as views)
    noise_level: variance of the residuals (1/tau);
                 per feature it is multiplied by a uniform random number in [0.5, 1.5] to model differences in features' noise
    scales, lscales: hyperparameters of the GP per factor (length as given by K)
    sample_cov: sample_covariates to use (can be of shape N X C) or "equidistant" or None
    shared: A list or single boolean indicating for each factor whether it is perfectly shared across groups or not.
    For non-shared ones pairwise group-group correlations are simulated by a Bernoulli distribution.
    Only relevant for factors with lengthscale and scale > 0.
    plot: If True, simulation results are plotted
    alpha: Pre-set activity for different views per factor. Needs to be a list of length M with arrays of length K. Default is None and will be drawn at random.
    """

    # simulate some test data
    np.random.seed(seed)
    M = len(views)
    N = int(N)
    if type(shared) == bool:
        shared = [shared]
    if len(shared) == 1:
        shared = [shared] * K

    groupidx = np.repeat(range(G), N) # kronecker structure
    if not sample_cov is None:
        if sample_cov == "equidistant":
            sample_cov = np.linspace(0,1,N)
            sample_cov = sample_cov.reshape(N, 1)
        else:
            assert sample_cov.shape[0] == N, "Number of rows of sample_cov and N does not match"
            if len(np.repeat(np.arange(0,100,1),2).shape) == 1:
                sample_cov = sample_cov.reshape(N, 1)
        distC = SS.distance.pdist(sample_cov, 'euclidean')**2.
        distC = SS.distance.squareform(distC)

    else:
        lscales = [0]* K

    Gmats = []
    for k in range(K):
        if scales[k] == 0 or lscales[k] == 0: # group structure not modelled
            Gmat = np.eye(G)
        else:
            if shared[k]:
                Gmat = np.ones([G,G])
            else:
                x = np.random.uniform(-1,1, G)
                Gmat = np.outer(x,x) + 0.5 * np.eye(G)
                Gmat = covar_to_corr(Gmat)
        Gmats.append(Gmat)

    # simulate Sigma
    Sigma =[]
    for k in range(K):
        if lscales[k] > 0:
            Kmat = scales[k] * np.exp(-distC / (2 * lscales[k] ** 2))
            Kmat = np.kron(Gmats[k], Kmat)
            Sigma.append( Kmat + (1-scales[k]) * np.eye(N*G))
        elif lscales[k] == 0:
            Kmat = scales[k] * (distC == 0).astype(float)
            Kmat = np.kron(Gmats[k], Kmat)
            Sigma.append(Kmat + (1-scales[k]) * np.eye(N*G))
            # Sigma.append(np.eye(N*G))
        else:
            sys.exit("All lengthscales need to be non-negative")

    # plot covariance structure
    if plot:
        fig, axs = plt.subplots(1, K, sharex=True, sharey=True)
        for k in range(K):
            sns.heatmap(Sigma[k], ax =axs[k])

    # simulate factor values
    Zks = []
    for k in range(K):
        sig = Sigma[k]
        Zks.append(np.random.multivariate_normal(np.zeros(N * G), sig, 1))
    Zks = np.vstack(Zks).transpose()

    Z = []
    for g in range(G):
        Z.append(Zks[groupidx == g,])

    # simulate alpha and theta, each factor should be active in at least one view
    theta = 0.5 * np.ones([M, K])
    if alpha is None:
        inactive = 1000
        active = 1
        alpha_tmp = [np.ones(M) * inactive]*K
        for k in range(K):
            while np.all(alpha_tmp[k]==inactive):
                alpha_tmp[k] = np.random.choice([active,inactive], size=M, replace=True)
        alpha = [ np.array(alpha_tmp)[:,m] for m in range(M) ]
    else:
        assert len(alpha) == M
        assert len(alpha[0]) == K

    # simulate weights
    W = []
    for m in range(M):
        W.append(np.column_stack(
            [np.random.normal(0, np.sqrt(1/alpha[m][k]), D[m]) * np.random.binomial(1, theta[m][k], D[m]) for k in range(K)]))

    # simulate heteroscedastic noise
    noise = []
    for m in range(M):
        tau_m = stats.uniform.rvs(loc=0.5, scale=1, size=D[m]) * 1/noise_level # uniform between 0.5 and 1.5 scaled by noise level
        noise.append(np.random.multivariate_normal(np.zeros(D[m]), np.eye(D[m]) * 1 / tau_m, N))

    # generate data
    data = []
    for m in range(M):
        tmp = []
        for g in range(G):
            tmp.append(Z[g].dot(W[m].transpose()) + noise[m])
        data.append(tmp)

    # store as list of groups
    if not sample_cov is None:
        sample_cov = [sample_cov] * G

    return {'data': data, 'W': W, 'Z': Z, 'noise': noise, 'sample_cov': sample_cov, 'Sigma': Sigma,
            'views': views, 'lscales': lscales, 'N': N, 'Gmats' : Gmats }



def mask_samples(sim, perc = 0.2, perc_all_views = 0):
    """
    Function to mask values at randomly sampled time points in each group and view.

    Param:
    perc: this fraction of time points are drawn in each view and group independently and all feature values at this time point set to NaN
    perc_all_views: this fraction of time points are drawn in each group independently and all feature values in all views are set to NaN
    """

    data = sim['data']
    N = sim['N']
    M = len(sim['views'])
    G = len(sim['data'][0])
    masked_samples = [[np.random.choice(N, math.floor(N * perc), replace = False) for g in range(G)] for m in range(M)]
    for m in range(M):
        for g in range(G):
            data[m][g][masked_samples[m][g],:] = s.nan

    if perc_all_views > 0:
        masked_samples.all_views = [np.random.choice(N, math.floor(N * perc_all_views), replace = False) for g in range(G)]
        for m in range(len(data)):
            for g in range(G):
                data[m][g][masked_samples.all_views[g], :] = s.nan


    return data



import pandas as pd
import torch
import numpy as np
import anndata
import matplotlib.pyplot as plt
from ._utils import load_data, data_process,bulk2single_data_prepare
from ._vae import train_vae, generate_vae, load_vae
#from .map_utils import create_data, DFRunner, joint_analysis, knn
import os
import warnings
import matplotlib
from typing import Union,Tuple
warnings.filterwarnings("ignore")


class Bulk2Single:
    r"""
    Bulk2Single class.
    
    """
    def __init__(self,bulk_data:pd.DataFrame,single_data:anndata.AnnData,celltype_key:str,
                 top_marker_num:int=500,ratio_num:int=1,gpu:Union[int,str]=0):
        """
        Initializes the Bulk2Single class.

        Arguments:
            bulk_data: The bulk RNA-seq data.
            single_data: The single-cell RNA-seq data.
            celltype_key: The name of the column in the bulk data containing cell types.
            top_marker_num: The number of top markers to select per cell type. Default is 500.
            ratio_num: The ratio between the number of single cells and target number of converted cells. Default is 1.
            gpu: The ID of the GPU to use. Set to -1 to use CPU. Default is 0. If set to 'mps', the MPS backend will be used.

        """
        self.bulk_data=bulk_data
        self.single_data=single_data
        self.celltype_key=celltype_key
        self.input_data=bulk2single_data_prepare(bulk_data,single_data,celltype_key)
        self.cell_target_num = data_process(self.input_data, top_marker_num, ratio_num)
        if gpu=='mps' and torch.backends.mps.is_available():
            print('Note that mps may loss will be nan, used it when torch is supported')
            self.used_device = torch.device("mps")
        else:
            self.used_device = torch.device(f"cuda:{gpu}") if gpu >= 0 and torch.cuda.is_available() else torch.device('cpu')
        self.history=[]

    def train(self,
            vae_save_dir:str='save_model',
            vae_save_name:str='vae',
            generate_save_dir:str='output',
            generate_save_name:str='output',
            batch_size:int=512,
            learning_rate:int=1e-4,
            hidden_size:int=256,
            epoch_num:int=5000,
            patience:int=50,save:bool=True)->torch.nn.Module:
        """
        Trains the VAE model.

        Arguments:
            vae_save_dir: The directory to save the trained VAE model. Default is 'save_model'.
            vae_save_name: The name of the saved VAE model. Default is 'vae'.
            generate_save_dir: The directory to save the generated single-cell data. Default is 'output'.
            generate_save_name: The name of the saved generated single-cell data. Default is 'output'.
            batch_size: The batch size for training. Default is 512.
            learning_rate: The learning rate for training. Default is 1e-4.
            hidden_size: The hidden size for the encoder and decoder networks. Default is 256.
            epoch_num: The maximum number of epochs for training. Default is 5.
            patience: The number of epochs to wait before early stopping. Default is 50.
            save: Whether to save the trained VAE model. Default is True.

        Returns:
            vae_net: The trained VAE model.
        """
        single_cell, label, breed_2_list, index_2_gene, cell_number_target_num, \
        nclass, ntrain, feature_size = self.__get_model_input(self.input_data, self.cell_target_num)
        print('...begin vae training')
        vae_net,history = train_vae(single_cell,
                            label,
                            self.used_device,
                            batch_size,
                            feature_size=feature_size,
                            epoch_num=epoch_num,
                            learning_rate=learning_rate,
                            hidden_size=hidden_size,
                            patience=patience,)
        print('...vae training done!')
        if save:
            path_save = os.path.join(vae_save_dir, f"{vae_save_name}.pth")
            if not os.path.exists(vae_save_dir):
                os.makedirs(vae_save_dir)
            torch.save(vae_net.state_dict(), path_save)
            print(f"...save trained vae in {path_save}.")
        self.vae_net=vae_net
        self.history=history
        return vae_net
        print('generating....')
        generate_sc_meta, generate_sc_data = generate_vae(vae_net, -1,
                                                          single_cell, label, breed_2_list,
                                                          index_2_gene, cell_number_target_num, used_device)
        sc_g=anndata.AnnData(generate_sc_data.T)
        sc_g.obs[self.celltype_key] = generate_sc_meta.loc[sc_g.obs.index,self.celltype_key].values
        sc_g.write_h5ad(os.path.join(generate_save_dir, f"{generate_save_name}.h5ad"), compression='gzip')
        self.__save_generation(generate_sc_meta, generate_sc_data,
                               generate_save_dir, generate_save_name)
        return sc_g
    
    def save(self,vae_save_dir:str='save_model',
            vae_save_name:str='vae',):
        """
        Saves the trained VAE model.

        Arguments:
            vae_save_dir: the directory to save the trained VAE model. Default is 'save_model'.
            vae_save_name: the name of the saved VAE model. Default is 'vae'.

        """
        path_save = os.path.join(vae_save_dir, f"{vae_save_name}.pth")
        if not os.path.exists(vae_save_dir):
            os.makedirs(vae_save_dir)
        torch.save(self.vae_net.state_dict(), path_save)
        print(f"...save trained vae in {path_save}.")
    
    def generate(self)->anndata.AnnData:
        r"""
        Generate the single-cell data.

        Returns:
            sc_g: The generated single-cell data.
        """
        single_cell, label, breed_2_list, index_2_gene, cell_number_target_num, \
        nclass, ntrain, feature_size = self.__get_model_input(self.input_data, self.cell_target_num)
        print('...generating')
        generate_sc_meta, generate_sc_data = generate_vae(self.vae_net, -1,
                                                          single_cell, label, breed_2_list,
                                                          index_2_gene, cell_number_target_num, self.used_device)
        generate_sc_meta.set_index('Cell',inplace=True)
        sc_g=anndata.AnnData(generate_sc_data.T)
        sc_g.obs[self.celltype_key] = generate_sc_meta.loc[sc_g.obs.index,'Cell_type'].values
        return sc_g
    
    def load(self,vae_load_dir:str,hidden_size:int=256):
        r"""
        load the trained VAE model of Bulk2Single.

        Arguments:
            vae_load_dir: The directory to load the trained VAE model.
            hidden_size: The hidden size for the encoder and decoder networks. Default is 256.
        """
        single_cell, label, breed_2_list, index_2_gene, cell_number_target_num, \
        nclass, ntrain, feature_size = self.__get_model_input(self.input_data, self.cell_target_num)
        print(f'loading model from {vae_load_dir}')
        vae_net = load_vae(feature_size, hidden_size, vae_load_dir, self.used_device)
        self.vae_net=vae_net

    def load_and_generate(self,
                              vae_load_dir:str,  # load_dir
                              hidden_size:int=256)->anndata.AnnData:
        r"""
        load the trained VAE model of Bulk2Single and generate the single-cell data.

        Arguments:
            vae_load_dir: The directory to load the trained VAE model.
            hidden_size: The hidden size for the encoder and decoder networks. Default is 256.

        Returns:
            sc_g: The generated single-cell data.
        """
        single_cell, label, breed_2_list, index_2_gene, cell_number_target_num, \
        nclass, ntrain, feature_size = self.__get_model_input(self.input_data, self.cell_target_num)
        print(f'loading model from {vae_load_dir}')
        vae_net = load_vae(feature_size, hidden_size, vae_load_dir, self.used_device)
        print('...generating')
        generate_sc_meta, generate_sc_data = generate_vae(vae_net, -1,
                                                          single_cell, label, breed_2_list,
                                                          index_2_gene, cell_number_target_num, self.used_device)
        generate_sc_meta.set_index('Cell',inplace=True)
        #return generate_sc_meta, generate_sc_data
        sc_g=anndata.AnnData(generate_sc_data.T)
        sc_g.obs[self.celltype_key] = generate_sc_meta.loc[sc_g.obs.index,'Cell_type'].values
        
        print('...generating done!')
        return sc_g
    
    def plot_loss(self,figsize:tuple=(4,4))->Tuple[matplotlib.figure.Figure,matplotlib.axes._axes.Axes]:
        r"""
        plot the loss curve of the trained VAE model.

        Arguments:
            figsize: The size of the figure. Default is (4,4).

        Returns:
            fig: The figure of the loss curve.
            ax: The axes of the figure.
        """
        fig, ax = plt.subplots(figsize=figsize)
        ax.plot(range(len(self.history)),self.history)
        ax.set_title('Beta-VAE')
        ax.set_ylabel('Loss')
        ax.set_xlabel('Epochs')
        return fig,ax


    def __get_model_input(self, data, cell_target_num):
        # input：data， celltype， bulk & output: label, dic, single_cell
        single_cell = data["input_sc_data"].values.T  # single cell data (600 * 6588)
        index_2_gene = (data["input_sc_data"].index).tolist()
        breed = data["input_sc_meta"]['Cell_type']
        breed_np = breed.values
        breed_set = set(breed_np)
        breed_2_list = list(breed_set)
        dic = {}  # breed_set to index {'B cell': 0, 'Monocyte': 1, 'Dendritic cell': 2, 'T cell': 3}
        label = []  # the label of cell (with index correspond)
        nclass = len(breed_set)

        ntrain = single_cell.shape[0]
        # FeaSize = single_cell.shape[1]
        feature_size = single_cell.shape[1]
        assert nclass == len(cell_target_num.keys()), "cell type num no match!!!"

        for i in range(len(breed_set)):
            dic[breed_2_list[i]] = i
        cell = data["input_sc_meta"]["Cell"].values

        for i in range(cell.shape[0]):
            label.append(dic[breed_np[i]])

        label = np.array(label)

        # label index the data size of corresponding target
        cell_number_target_num = {}
        for k, v in cell_target_num.items():
            cell_number_target_num[dic[k]] = v

        return single_cell, label, breed_2_list, index_2_gene, cell_number_target_num, nclass, ntrain, feature_size

    def __save_generation(self, generate_sc_meta, generate_sc_data, generate_save_dir,
                          generate_save_name, ):
        # saving.....
        if not os.path.exists(generate_save_dir):
            os.makedirs(generate_save_dir)
        path_label_generate_csv = os.path.join(generate_save_dir, f"{generate_save_name}_sc_celltype.csv")
        path_cell_generate_csv = os.path.join(generate_save_dir, f"{generate_save_name}_sc_data.csv")

        generate_sc_meta.to_csv(path_label_generate_csv)
        generate_sc_data.to_csv(path_cell_generate_csv)
        print(f"saving to {path_label_generate_csv} and {path_cell_generate_csv}.")



import numpy as np
import warnings

from ..utils import coms_list_to_matrix

__all__ = [
    'symmetric_jaccard',
    'overlapping_nmi',
]


def symmetric_jaccard(coms_1, coms_2):
    """Quantify agreement between two community assignments based on symmetric Jaccard similarity.

    Computed as in the CESNA paper as
    0.5 * (1 / |C1| * sum_{c1 in C1} max_{c2 in C2} jac(c1, c2) +
           1 / |C2| * sum_{c2 in C2} max_{c1 in C1} jac(c1, c2))

    Parameters
    ----------
    coms_1 : list of len [num_coms] or array-like of shape [num_nodes, num_coms]
        First community assignment to compare.
    coms_2 : list of len [num_coms] or array-like of shape [num_nodes, num_coms]
        Second community assignment to compare.

    Returns
    -------
    symmetric_jaccard_similarity : float
        Symmetric average best Jaccard similarity between two community assignments.

    """
    # Convert community assignments to matrix format
    if isinstance(coms_1, list):
        F1 = coms_list_to_matrix(coms_1)
    elif len(coms_1.shape) == 2:
        F1 = coms_1
    else:
        raise ValueError("coms_1 must be either a list or a matrix.")
    if isinstance(coms_2, list):
        F2 = coms_list_to_matrix(coms_2)
    elif len(coms_2.shape) == 2:
        F2 = coms_2
    else:
        raise ValueError("coms_2 must be either a list or a matrix.")

    intersections = F1.T.dot(F2)
    sum_F1 = F1.sum(0)
    sum_F2 = F2.sum(0)
    unions = (np.ravel(sum_F2) + np.ravel(sum_F1)[:, None]) - intersections
    jacs = intersections / unions
    return 0.5 * (jacs.max(0).mean() + jacs.max(1).mean())


def overlapping_nmi(X, Y):
    """Compute NMI between two overlapping community covers.

    Parameters
    ----------
    X : array-like, shape [N, m]
        Matrix with samples stored as columns.
    Y : array-like, shape [N, n]
        Matrix with samples stored as columns.

    Returns
    -------
    nmi : float
        Float in [0, 1] quantifying the agreement between the two partitions.
        Higher is better.

    References
    ----------
    McDaid, Aaron F., Derek Greene, and Neil Hurley.
    "Normalized mutual information to evaluate overlapping
    community finding algorithms."
    arXiv preprint arXiv:1110.2515 (2011).

    """
    if not ((X == 0) | (X == 1)).all():
        raise ValueError("X should be a binary matrix")
    if not ((Y == 0) | (Y == 1)).all():
        raise ValueError("Y should be a binary matrix")

    if X.shape[1] > X.shape[0] or Y.shape[1] > Y.shape[0]:
        warnings.warn("It seems that you forgot to transpose the F matrix")
    X = X.T
    Y = Y.T
    def cmp(x, y):
        """Compare two binary vectors."""
        a = (1 - x).dot(1 - y)
        d = x.dot(y)
        c = (1 - y).dot(x)
        b = (1 - x).dot(y)
        return a, b, c, d

    def h(w, n):
        """Compute contribution of a single term to the entropy."""
        if w == 0:
            return 0
        else:
            return -w * np.log2(w / n)

    def H(x, y):
        """Compute conditional entropy between two vectors."""
        a, b, c, d = cmp(x, y)
        n = len(x)
        if h(a, n) + h(d, n) >= h(b, n) + h(c, n):
            return h(a, n) + h(b, n) + h(c, n) + h(d, n) - h(b + d, n) - h(a + c, n)
        else:
            return h(c + d, n) + h(a + b, n)
    def H_uncond(X):
        """Compute unconditional entropy of a single binary matrix."""
        return sum(h(x.sum(), len(x)) + h(len(x) - x.sum(), len(x)) for x in X)

    def H_cond(X, Y):
        """Compute conditional entropy between two binary matrices."""
        m, n = X.shape[0], Y.shape[0]
        scores = np.zeros([m, n])
        for i in range(m):
            for j in range(n):
                scores[i, j] = H(X[i], Y[j])
        return scores.min(axis=1).sum()

    if X.shape[1] != Y.shape[1]:
        raise ValueError("Dimensions of X and Y don't match. (Samples must be stored as COLUMNS)")
    H_X = H_uncond(X)
    H_Y = H_uncond(Y)
    I_XY = 0.5 * (H_X + H_Y - H_cond(X, Y) - H_cond(Y, X))
    return I_XY / max(H_X, H_Y)



import numpy as np
import torch

from copy import deepcopy


class ModelSaver:
    """In-memory saver for model parameters.

    Storing weights in memory is faster than saving to disk with torch.save.
    """
    def __init__(self, model):
        self.model = model

    def save(self):
        self.state_dict = deepcopy(self.model.state_dict())

    def restore(self):
        self.model.load_state_dict(self.state_dict)


class EarlyStopping:
    """Base class for an early stopping monitor that says when it's time to stop training.

    Examples
    --------
    early_stopping = EarlyStopping()
    for epoch in range(max_epochs):
        sess.run(train_op)  # perform training operation
        early_stopping.next_step()
        if early_stopping.should_stop():
            break
        if early_stopping.should_save():
            model_saver.save()  # save model weights

    """
    def __init__(self):
        pass

    def reset(self):
        """Reset the internal state."""
        raise NotImplementedError

    def next_step(self):
        """Should be called at every iteration."""
        raise NotImplementedError

    def should_save(self):
        """Says if it's time to save model weights."""
        raise NotImplementedError

    def should_stop(self):
        """Says if it's time to stop training."""
        raise NotImplementedError


class NoEarlyStopping(EarlyStopping):
    """No early stopping."""
    def __init__(self):
        super().__init__()
        pass

    def reset(self):
        pass

    def next_step(self):
        pass

    def should_stop(self):
        return False

    def should_save(self):
        return False


class NoImprovementStopping(EarlyStopping):
    """Stop training when the validation metric stops improving.

    Parameters
    ----------
    validation_fn : function
        Calling this function returns the current value of the validation metric.
    mode : {'min', 'max'}
        Should the validation metric be minimized or maximized?
    patience : int
        Number of iterations without improvement before stopping.
    tolerance : float
        Minimal improvement in validation metric to not trigger patience.
    relative : bool
        Is tolerance measured in absolute units or relatively?

    Attributes
    ----------
    _best_value : float
        Best value of the validation loss.
    _num_bad_epochs : int
        Number of epochs since last significant improvement in validation metric.
    _time_to_save : bool
        Is it time to save the model weights?
    _is_better : function
        Tells if new validation metric value is better than the best one so far.
        Signature self._is_better(new_value, best_value).

    """
    def __init__(self, validation_fn, mode='min', patience=10, tolerance=0.0, relative=False):
        super().__init__()
        self.validation_fn = validation_fn
        self.mode = mode
        self.patience = patience
        self.tolerance = tolerance
        self.relative = relative
        self.reset()

        if mode not in ['min', 'max']:
            raise ValueError(f"Mode should be either 'min' or 'max' (got {mode} instead).")

        # Create the comparison function
        if relative:
            if mode == 'min':
                self._is_better = lambda new, best: new < best - (best * tolerance)
            if mode == 'max':
                self._is_better = lambda new, best: new > best + (best * tolerance)
        else:
            if mode == 'min':
                self._is_better = lambda new, best: new < best - tolerance
            if mode == 'max':
                self._is_better = lambda new, best: new > best + tolerance

    def reset(self):
        """Reset the internal state."""
        self._best_value = self.validation_fn()
        self._num_bad_epochs = 0
        self._time_to_save = False

    def next_step(self):
        """Should be called at every iteration."""
        last_value = self.validation_fn()
        if self._is_better(last_value, self._best_value):
            self._time_to_save = True
            self._best_value = last_value
            self._num_bad_epochs = 0
        else:
            self._num_bad_epochs += 1

    def should_save(self):
        """Says if it's time to save model weights."""
        if self._time_to_save:
            self._time_to_save = False
            return True
        else:
            return False

    def should_stop(self):
        """Says if it's time to stop training."""
        return self._num_bad_epochs > self.patience



from __future__ import division
from .. import gpu_utils
from ..distributions import *
import scipy as s
import numpy as np

# Import manually defined functions
from .variational_nodes import MultivariateGaussian_Unobserved_Variational_Node

# TODO:
# - integrate into Z node using ix

# U_GP_Node_mv
class U_GP_Node_mv(MultivariateGaussian_Unobserved_Variational_Node):
    """
    U node with a Multivariate Gaussian prior and variational distribution
    """

    def __init__(self, dim, pmean, pcov, qmean, qcov, qE=None, idx_inducing = None, weight_views = False):
        super().__init__(dim=dim, pmean=pmean, pcov=pcov, qmean=qmean, qcov=qcov, qE=qE)

        self.mini_batch = None
        self.factors_axis = 1
        self.Nu = self.dim[0]
        self.K = self.dim[1]
        self.idx_inducing = idx_inducing
        self.weight_views = weight_views

        assert len(self.idx_inducing) == self.Nu, "Dimension of U and number of inducing points does not match"

        # Precompute terms (inverse covariance ant its determinant for each factor) to speed up computation
        # self.p_cov = self.P.params['cov']
        # self.p_cov_inv = np.array([s.linalg.inv(cov) for cov in tmp])
        # self.p_cov_inv_diag = np.array([np.diag(c) for c in self.p_cov_inv])

    def precompute(self, options):
        """ Method to precompute terms to speed up computation """
        gpu_utils.gpu_mode = options['gpu_mode']

    def removeFactors(self, idx, axis=0):
        super().removeFactors(idx, axis)
        self.K = self.dim[1]
        # self.p_cov = np.delete(self.p_cov, axis=0, obj=idx)
        # self.p_cov_inv = np.delete(self.p_cov_inv, axis=0, obj=idx)
        # self.p_cov_inv_diag = np.delete(self.p_cov_inv_diag, axis=0, obj=idx)


    def get_mini_batch(self):
        """ Method to fetch minibatch """
        if self.mini_batch is None:
            return self.getExpectations()
        else:
            return self.mini_batch

    def updateParameters(self, ix=None, ro=1.):

        # Get expectations from other nodes
        W = self.markov_blanket["W"].getExpectations()
        Y = self.markov_blanket["Y"].get_mini_batch()
        tau = self.markov_blanket["Tau"].get_mini_batch()
        Z = self.markov_blanket["Z"].get_mini_batch()
        mask = [self.markov_blanket["Y"].nodes[m].getMask() for m in range(len(Y))]

        assert "Sigma" in self.markov_blanket, "Sigma not found in Markov blanket of U node"
        Sigma = self.markov_blanket['Sigma'].get_mini_batch()
        SigmaUZ = Sigma['cov'][:,self.idx_inducing, :]
        p_cov_inv = Sigma['inv']


        # Get variational parameters of current node
        Q = self.Q.getParameters()
        Qmean, Qcov = Q['mean'], Q['cov']

        par_up = self._updateParameters(Y, W, Z, tau, Qmean, Qcov, SigmaUZ, p_cov_inv, mask)

        # Update parameters
        Q['mean'] = par_up['Qmean']
        Q['cov'] = par_up['Qcov']

        self.Q.setParameters(mean=Q['mean'], cov=Q['cov'])  # NOTE should not be necessary but safer to keep for now

    def _updateParameters(self, Y, W, Z, tau, Qmean, Qcov, SigmaUZ, p_cov_inv, mask):
        """ Hidden method to compute parameter updates """

        M = len(Y)
        N = Y[0].shape[0]
        K = self.dim[1]

        # Masking
        for m in range(M):
            tau[m][mask[m]] = 0.

        weights = [1] * M
        if self.weight_views and M > 1:
            total_w = np.asarray([Y[m].shape[1] for m in range(M)]).sum()
            weights = np.asarray([total_w / (M * Y[m].shape[1]) for m in range(M)])
            weights = weights / weights.sum() * M

        # Precompute terms to speed up GPU computation
        foo = gpu_utils.array(np.zeros((N, K)))
        precomputed_bar = gpu_utils.array(np.zeros((N, K)))
        for m in range(M):
            tau_gpu = gpu_utils.array(tau[m])
            foo += weights[m] * gpu_utils.dot(tau_gpu, gpu_utils.array(W[m]["E2"]))
            bar_tmp1 = gpu_utils.array(W[m]["E"])
            bar_tmp2 = tau_gpu * gpu_utils.array(Y[m])
            precomputed_bar += weights[m] * gpu_utils.dot(bar_tmp2, bar_tmp1)
        foo = gpu_utils.asnumpy(foo)

        # Calculate variational updates - term for mean
        for k in range(K):
            bar = gpu_utils.array(np.zeros((N,)))
            tmp_cp1 = gpu_utils.array(Z['E'][:, np.arange(K) != k])
            for m in range(M):
                tmp_cp2 = gpu_utils.array(W[m]["E"][:, np.arange(K) != k].T)

                bar_tmp1 = gpu_utils.array(W[m]["E"][:, k])
                bar_tmp2 = gpu_utils.array(tau[m]) * (-gpu_utils.dot(tmp_cp1, tmp_cp2))

                bar += weights[m] * gpu_utils.dot(bar_tmp2, bar_tmp1)
            bar += precomputed_bar[:, k]
            bar = gpu_utils.asnumpy(bar)

            # note: no Alpha scaling required here compared to Z nodes as done in the updateParameters function
            Mcross = gpu_utils.dot(p_cov_inv[k, :, :], SigmaUZ[k, :, :])
            Mtmp = gpu_utils.dot(Mcross, gpu_utils.dot(np.diag(foo[:, k]), Mcross.transpose()))
            Qcov[k, :, :] = np.linalg.inv(Mtmp + p_cov_inv[k, :, :])
            Qmean[:, k] = gpu_utils.dot(Qcov[k, :, :], gpu_utils.dot(gpu_utils.dot(p_cov_inv[k, :, :], SigmaUZ[k, :, :]) , bar))

        return {'Qmean': Qmean, 'Qcov': Qcov}

    def calcELBOgrad_k(self, k, gradSigma):
        """
        Method to calculate ELBO gradients per factor - required for optimization in Sigma node
        """
        Qpar, Qexp = self.Q.getParameters(), self.Q.getExpectations()
        Qmean, Qcov = Qpar['mean'], Qpar['cov']
        QE = Qexp['E']

        assert "Sigma" in self.markov_blanket, "Sigma not found in Markov blanket of U node"
        Sigma = self.markov_blanket['Sigma'].getExpectations()
        p_cov = Sigma['cov']
        p_cov_inv = Sigma['inv']
        p_cov_inv_logdet = Sigma['inv_logdet']

        term1 = - 0.5 * np.trace(gpu_utils.dot(gradSigma, p_cov_inv[k, :,:]))
        term2 = 0.5 * np.trace(gpu_utils.dot(p_cov_inv[k, :,:], gpu_utils.dot(gradSigma, gpu_utils.dot(p_cov_inv[k, :,:],  Qcov[k, :, :]))))
        term3 = 0.5 * gpu_utils.dot(QE[:, k].transpose(), gpu_utils.dot(p_cov_inv[k, :,:], gpu_utils.dot(gradSigma, gpu_utils.dot(p_cov_inv[k, :,:], QE[:,k]))))

        return term1 + term2 + term3

    # Eblo calculation per factor - required for grid search for optimal hyperparameter in sigma per factor
    def calculateELBO_k(self, k):
        # Collect parameters and expectations of current node
        Qpar, Qexp = self.Q.getParameters(), self.Q.getExpectations()
        Qmean, Qcov = Qpar['mean'], Qpar['cov']

        QE = Qexp['E']

        assert "Sigma" in self.markov_blanket, "Sigma not found in Markov blanket of U node"

        Sigma = self.markov_blanket['Sigma'].getExpectations()
        p_cov = Sigma['cov']
        p_cov_inv = Sigma['inv']
        p_cov_inv_logdet = Sigma['inv_logdet']

        # compute term from the exponential in the Gaussian
        tmp1 = -0.5 * (np.trace(gpu_utils.dot(p_cov_inv[k,:,:], Qcov[k, :, :])) + gpu_utils.dot(QE[:, k].transpose(),
                                                                                                  gpu_utils.dot(
                                                                                                      p_cov_inv[k,:,:], QE[:,k])))  # expectation of quadratic form

        # compute term from the precision factor in front of the Gaussian
        tmp2 = 0.5 * p_cov_inv_logdet[k]
        lb_p = tmp1 + tmp2

        lb_q = -0.5 * np.linalg.slogdet(Qcov[k, :, :])[1]

        # term -N*(log(2* np.pi)) cancels out between p and q term; -N/2 is added below
        return lb_p - lb_q

    # sum up individual ELBO calculations
    def calculateELBO(self):
        elbo = 0
        for k in range(self.dim[1]):
            elbo += self.calculateELBO_k(k)
        elbo -= self.dim[0] * self.dim[1] / 2.

        return elbo




"""
Module to train a bioFAM model
"""

import scipy as s
import pandas as pd
import numpy as np

from ..core.BayesNet import BayesNet


def train_model(model):

    # Sanity check on the Bayesian Network
    assert isinstance(model, BayesNet), "'model' has to be a BayesNet class"


    ####################
    ## Start training ##
    ####################

    print ("\n")
    print ("#"*38)
    print ("## Training the model with seed %d ##" % (model.options['seed']))
    print ("#"*38)
    print ("\n")

    model.iterate()

    print("\n")
    print("#"*23)
    print("## Training finished ##")
    print("#"*23)
    print("\n")



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd 
import networkx as nx
import itertools
from ..utils import plot_network
from typing import Union,Tuple
import matplotlib

def string_interaction(gene:list,species:int) -> pd.DataFrame:
    r"""A Python library to analysis the protein-protein interaction network by string-db
  
    Arguments:
        gene: The gene list to analysis PPI
        species: NCBI taxon identifiers (e.g. Human is 9606, see: STRING organisms).
    
    Returns:
        res: the dataframe of protein-protein interaction
    
    """
    import requests ## python -m pip install requests
    string_api_url = "https://string-db.org/api"
    output_format = "tsv-no-header"
    method = "network"
    request_url = "/".join([string_api_url, output_format, method])
    my_genes = gene

    params = {

        "identifiers" : "%0d".join(my_genes), # your protein
        "species" : species, # species NCBI identifier 
        "caller_identity" : "www.awesome_app.org" # your app name

    }
    response = requests.post(request_url, data=params)
    res=pd.DataFrame(columns=['stringId_A',\
                             'stringId_B',\
                             'preferredName_A',\
                             'preferredName_B',\
                             'ncbiTaxonId',\
                             'score',\
                             'nscore',\
                             'fscore',\
                             'pscore',\
                             'ascore',\
                             'escore',\
                             'dscore',\
                             'tscore'])
    num=0
    for line in response.text.strip().split("\n"):
        l = line.strip().split("\t")
        res.loc[num]={'stringId_A':l[0],\
                             'stringId_B':l[1],\
                             'preferredName_A':l[2],\
                             'preferredName_B':l[3],\
                             'ncbiTaxonId':l[4],\
                             'score':l[5],\
                             'nscore':l[6],\
                             'fscore':l[7],\
                             'pscore':l[8],\
                             'ascore':l[9],\
                             'escore':l[10],\
                             'dscore':l[11],\
                             'tscore':l[12]}
        num+=1
    return res



def string_map(gene:list,species:int)->pd.DataFrame:
    r"""A Python library to find the gene name in string-db

    Arguments:
        gene: The gene list to analysis PPI
        species: NCBI taxon identifiers (e.g. Human is 9606, see: STRING organisms).
    
    Returns:
        res: the dataframe of query gene and new gene
    
    """
    import requests ## python -m pip install requests
    string_api_url = "https://string-db.org/api"
    output_format = "tsv-no-header"
    method = "get_string_ids"
    params = {

        "identifiers" : "\r".join(gene), # your protein list
        "species" : species, # species NCBI identifier 
        "limit" : 1, # only one (best) identifier per input protein
        "echo_query" : 1, # see your input identifiers in the output
        "caller_identity" : "www.awesome_app.org" # your app name

    }
    request_url = "/".join([string_api_url, output_format, method])
    results = requests.post(request_url, data=params)
    res=pd.DataFrame(columns=['queryItem','queryIndex','stringId','ncbiTaxonId','taxonName','preferredName','annotation'])
    num=0
    for line in results.text.strip().split("\n"):
        l = line.split("\t")
        res.loc[num]={'queryItem':l[0],\
          'queryIndex':l[1],\
          'stringId':l[2],\
          'ncbiTaxonId':l[3],\
          'taxonName':l[4],\
          'preferredName':l[5],\
          'annotation':l[6]}
        num+=1
    return res


def max_interaction(gene,species):
    gene_len=len(gene)
    times=gene_len//1000
    shengyu=gene_len-times*1000
    if shengyu!=0:
        times+=1
    ge=[]
    for i in range(times):
        ge.append(gene[1000*i:1000*(i+1)])
    b=[]
    for p in itertools.combinations(range(times),2):
        b.append(string_interaction(ge[p[0]]+ge[p[1]],species))
    res=pd.concat(b,axis=0,ignore_index=True)
    res=res.drop_duplicates()
    return res
    

def generate_G(gene:list,species:int,score:float=0.4) -> nx.Graph:
    r"""A Python library to get the PPI network in string-db

    Arguments:
        gene: The gene list to analysis PPI
        species: NCBI taxon identifiers (e.g. Human is 9606, see: STRING organisms).
        score: The threshold of protein A and B interaction

    Returns:
        G: the networkx object of PPI in query gene list
    
    """
    
    a=string_interaction(gene,species)
    b=a.drop_duplicates()
    b.head()
    G = nx.Graph()
    G.add_nodes_from(set(b['preferredName_A'].tolist()+b['preferredName_B'].tolist()))

    #Connect nodes
    for i in b.index:
        col_label = b.loc[i]['preferredName_A']
        row_label = b.loc[i]['preferredName_B']
        if(float(b.loc[i]['score'])>score):
            G.add_edge(col_label,row_label)
    return G

class pyPPI(object):

    def __init__(self,gene: list,species: int,gene_type_dict: dict,gene_color_dict: dict,
                 score: float = 0.4) -> None:
        """Initialize the protein-protein interaction analysis.

        Arguments:
            gene: The gene list to analysis PPI
            species: NCBI taxon identifiers (e.g. Human is 9606, see: STRING organisms).
            gene_type_dict: The gene type dict, the key is gene name, the value is gene type.
            gene_color_dict: The gene color dict, the key is gene name, the value is gene color.
            score: The threshold of protein A and B interaction
        """
        self.gene=gene 
        self.species=species
        self.score=score
        self.gene_type_dict=gene_type_dict
        self.gene_color_dict=gene_color_dict

    def interaction_analysis(self) -> nx.Graph:
        """Analysis the protein-protein interaction.
        
        Returns:
            G: the networkx object of PPI in query gene list
        """
        G=generate_G(self.gene,
                          self.species,
                          self.score)
        self.G=G 
        return G
    
    def plot_network(self,**kwargs) -> Tuple[matplotlib.figure.Figure,matplotlib.axes._axes.Axes]:
        """Plot the protein-protein interaction network.

        Returns:
            fig: the figure of PPI network
            ax: the AxesSubplot of PPI network
        """
        return plot_network(self.G,self.gene_type_dict,self.gene_color_dict,**kwargs)
    
    
        


from .Alpha_nodes import AlphaW_Node, AlphaZ_Node
from .Tau_nodes import TauD_Node
from .Theta_nodes import ThetaW_Node, ThetaZ_Node
from .W_nodes import W_Node, SW_Node
from .Y_nodes import Y_Node
from .Z_nodes import Z_Node, SZ_Node
from .Z_nodes_GP import Z_GP_Node
from .Z_nodes_GP_mv import Z_GP_Node_mv
from .multiview_nodes import *
from .basic_nodes import *
from .nongaussian_nodes import *
from .variational_nodes import *
from .Sigma_node import *
from .Kc_node import Kc_Node
from .Kg_node import Kg_Node
from .U_nodes import U_GP_Node_mv
from .ZgU_node import ZgU_node